from __future__ import print_function
import numpy as np
from matplotlib import pyplot as plt
import math
from scipy import interpolate as ip
from numpy import ma
import os
import pdb
from copy import deepcopy
from dataIO.span import span,span_from_pixels
from pySurf.plane_fit import plane_fit
from astropy.io import fits
#from matplotlib.mlab import griddata
#from scipy.interpolate import griddata
from dataIO.running_mean import running_mean
from scipy import stats
from plotting.add_clickable_markers import add_clickable_markers2


verbose = False

"""
Module containing functions acting on a point cloud. With the intention
of creating a class. Points are in format (Npoints,Ndim).
Note also that shape, when used (mostly is just a visual parameter for plots), is in format (nx,ny), that is opposite to python convention.
"""

method='linear' #used for ip.griddata (scipy.interpolate.griddata)
# note from scipy.interpolate import interp2d to interpolate from grid to grid
# also scipy.ndimage.map_coordinates

"""2014/03/17 moved here all routines from EA2019_align and from OP1S07_align."""

"""2014/03/06 from v1. Start from nanovea txt files use pure python to avoid coordinates mess. Assume data are saved with y fast axis and +,+ scanning
directions. It is not guaranteed to work with different scanning directions."""

"""angpos copiata qui, che non so come si fa ad importare in questo casino di directories.
CHANGES: angpos e' stata corretta funzionava con dati trasposti"""

## TRANSFORMATIONS
def crop_points1(points,xrange=None,yrange=None,zrange=None,
    poly=None,interactive=False):
    """crop a xyz points [Nx3], keeping only points inside xrange and yrange defined as (min,max).
        Interactive allows to set a region by zooming and/or a polygon by point and click
        (order of vertex matters)."""



    if interactive:
        """minimally tested."""

        curfig=plt.gcf() if plt.get_fignums() else None
        fig=plt.figure()
        plot_points(points,scatter=True)
        print ("""Zoom to the region to crop, and/or use CTRL+leftClick to add points and create
        an polygonal selection. CTRL+rightClick remove the nearest point. Press ENTER when done.""")
        ax=add_clickable_markers2(hold=True,propertyname='polypoints')
        if xrange is None:
            xrange=plt.xlim()
        if yrange is None:
            yrange=plt.ylim()
        mh=ax.polypoints
        del ax.polypoints
        #points[~points_in_poly(points,mh),2]=np.nan
        #points=points[~np.isnan(points[:,2]),:]
        points=points[points_in_poly(points,mh),:]

        plt.close(fig)
        if curfig:
            plt.figure(curfig.number);

    if poly:
        raise NotImplementedError

    if xrange is not None:
        if xrange[0] is not None:
            points=points[(xrange[0]<=points[:,0]),:]
        if xrange[1] is not None:
            points=points[(points[:,0]<=xrange[1]),:]
    if yrange is not None:
        if yrange[0] is not None:
            points=points[(yrange[0]<=points[:,1]),:]
        if yrange[1] is not None:
            points=points[(points[:,1]<=yrange[1]),:]
    if zrange is not None:
        if zrange[0] is not None:
            points=points[(zrange[0]<=points[:,2]),:]
        if zrange[1] is not None:
            points=points[(points[:,2]<=zrange[1]),:]
    return points

def crop_points(points,xrange=None,yrange=None,zrange=None,mask=False,poly=None,interactive=False):
    """experimental version, adds option booleam to return mask. useful e.g. to clean data based on crop on deltaR"""
    """crop a xyz points [Nx3], keeping only points inside xrange and yrange defined as (min,max)."""

    #outmask is sized as points and must remain same size all time.
    # at the end if flag mask is set, outmask is returned, else points are filtered.
    outmask=np.ones(points.shape[0]).astype(bool)
    #if mask is not None:
    #    outmask &= mask

    if interactive:
        """minimally tested."""

        curfig=plt.gcf() if plt.get_fignums() else None
        fig=plt.figure()
        plot_points(points,scatter=True)
        print ("""Zoom to the region to crop, and/or use CTRL+leftClick to add points and create
        an polygonal selection. CTRL+rightClick remove the nearest point. Press ENTER when done.""")
        ax=add_clickable_markers2(hold=True,propertyname='polypoints')
        if xrange is None:
            xrange=plt.xlim()
        if yrange is None:
            yrange=plt.ylim()
        mh=ax.polypoints
        del ax.polypoints
        poly=mh
        plt.close(fig)
        if curfig:
            plt.figure(curfig.number);

    if poly:
        outmask=outmask & points_in_poly(points,poly)

    if xrange is not None:
        if xrange[0] is not None:
            outmask &= (xrange[0]<=points[:,0])
        if xrange[1] is not None:
            outmask &= (points[:,0]<=xrange[1])
    if yrange is not None:
        if yrange[0] is not None:
            outmask &= (yrange[0]<=points[:,1])
        if yrange[1] is not None:
            outmask &= (points[:,1]<=yrange[1])
    if zrange is not None:
        if zrange[0] is not None:
            outmask &= (zrange[0]<=points[:,2])
        if zrange[1] is not None:
            outmask &= (points[:,2]<=zrange[1])

    points=outmask if mask else points[outmask]

    return points


def rotate_points(points,theta,center=(0,0)):
    """returns rotated coordinates of 2D point(s) x ([Npoints x 2]) about a center with anticlockwise angle theta in rad. If 3D points are passed, z coordinate is maintained."""
    tx,ty=center
    points=np.array(points)
    if (np.shape(points)[-1]==3):
        return np.hstack([rotate_points(points[:,0:2],theta,center),points[:,2:]])
    else:
        if(np.shape(points)[-1]!=2):
            raise ValueError
    x,y=np.array(points[:,0]),np.array(points[:,1])
    cost=math.cos(theta)
    sint=math.sin(theta)
    x1=x*cost-y*sint + ty*sint - tx*(cost-1) #+cost*tx #
    y1=x*sint+y*cost - tx*sint - ty *(cost-1) #-sint*tx #
    return np.vstack((x1,y1)).T

def translate_points(points,offset=None):
    """returns translated coordinates. Useless routine, can be done with
    a simple sum points+offset"""
    ndim=points.shape[-1]
    if offset is None:
        offset=np.zeros(ndim)
    else:
        assert (len(offset)==ndim)
    return points+offset

def _get_plane(points,pars=None,xrange=None,yrange=None,zrange=None,mask=None,returnPars=False):
    """Return points of a plane defined by pars on x,y coordinates of points.
    pars is a 3 elements vector [A,B,C] according to Ax + By + C = z.
    if pars is None, plane is calculated by best fit on selected points and two elements are returned:
    the points of the best fit plane and its parameters A,B,C. The points are calculated on all
    points positions."""
    #2016/02/11 modified to return only plane (and not pars) by default, as it was before it was
    #   returning a tuple (plane, pars) if and only if pars was not set.
    #   this breaks the interface, so I added the initial undersore in the function name, that it
    #   seems to be called only from inside points.py. If other programs are using the function
    #   they need to be adjusted by explicitly adding returnPars=True

    x,y,z=np.hsplit(points,3)
    if pars is None:
        if mask is not None:
            points=points[mask,:]
        pp=crop_points(points,xrange=xrange,yrange=yrange,zrange=zrange)
        pars=plane_fit(*np.hsplit(pp,3)).flatten()

    A,B,C=pars
    plane=np.hstack([x,y,A*x+B*y+C])
    if returnPars:
        return plane,pars
    else:
        return plane

def level_points(points,xrange=None,yrange=None,zrange=None,mask=None,pars=None,retall=False):
    """return the leveled points (after subtraction of best fit plane on selected points).
    If pars is provided as 3 elements vector is intended as plane coefficients and fit is not performed."""

    matrix=False
    if points.shape[1]!=3: #matrix input
        xgrid=np.arange(points.shape[0]) #irrelevant, they get lost when result is converted
        ygrid=np.arange(points.shape[1]) #   back to matrix.
        points=matrix_to_points(points,xgrid,ygrid)
        matrix=True  #flag for final resampling

    #get plane and pars, pars are used only if not passed (if they are passed there is no fit)
    plane=_get_plane(points,pars=pars,xrange=xrange,yrange=yrange,zrange=zrange,mask=mask,returnPars=True)
    #adjust for the number of variables needed.
    if pars is None:
        pars=plane[1]
        plane=plane[0]
    else:
        plane=plane[0]
    points=subtract_points(points,plane,resample=0)

    if matrix:
        points=resample_grid(points,xgrid,ygrid,matrix=True,resample=False)

    if retall:
        return points,pars
    else:
        return points
'''
def trim_points(points,interval,missing=None,invert=False):
    """Trim points keeping only points with z inside the interval.
    It differs from clip in the fact that points are eliminated, not replaced
    and clipping levels can be set only directly (as opposite as being calculated
    in automatic.
    But at the end is the same as crop. Use that."""

    aa=(points[:,2]<interval[1]) & (points[:,2]>interval[0]) #points to keep
    if invert:
       aa=np.logical_not(aa)
    na=np.logical_not(aa)   #negative of aa
    if missing is None:
        points=points[aa,:]
    else:
        points[na,2]=missing

    return points
'''


## ANALYSIS
def points_find_hull(pts):
    """return the convex hull non containing invalid (nan) points as a (np,2) array.
    hull=points_find_hull(pts)
    plt.plot(hull[:,0], hull[:,1], 'r--', lw=2)."""

    from scipy.spatial import ConvexHull
    if pts.shape[1] == 2:
        xypts = pts
    elif pts.shape[1] == 3:
        xypts=pts[~np.isnan(pts[:,2]),:2]
    else:
        raise ValueError("wrong shape for points in points_find_hull")

    cv = ConvexHull(xypts)

    ihull = cv.vertices
    # the vertices of the convex hull
    hull=xypts[ihull,:]

    return hull


def points_in_poly2(pts,vert):
    """return a boolen array, True if point is inside a polygon,
     given its vertices."""
    from shapely.geometry import Point
    from shapely.geometry.polygon import Polygon
    inpoly=np.zeros(pts.shape[0]).astype(bool)


    point = Point(0.5, 0.5)
    polygon = Polygon(verts)
    return polygon.contains(point)

def points_in_poly(pts,verts):
    from matplotlib import path
    p=path.Path(verts)
    if pts.shape[1] == 3:
        pts=pts[:,:2]
    elif pts.shape[1] != 2:
        raise ValueError("wrong shape for points in points_find_hull")
    return p.contains_points(pts)
    #return np.hstack([pts[:,:2],p.contains_points(pts)[:,np.newaxis]])



def clipStats(p,clip):
    print('clip for %s : %s'%(clip))
    print('z range: %s : %s'%(np.nanmin(p[:,2]),np.nanmax(p[:,2])))
    print('rms (clipped|unclipped): %s , %s'%(points_rms(p,clip=clip),points_rms(p)))
    print('number of elements clipped below|above %s,%s'%(p[np.where(p[:,2]<clip[0]),2].size,p[np.where(p[:,2]>clip[1]),2].size))
    print('xrange: %s,%s '%(np.nanmin(p[:,0]),np.nanmax(p[:,0])))
    print('yrange: %s,%s '%(np.nanmin(p[:,1]),np.nanmax(p[:,1])))

def points_rms(points,xrange=None,yrange=None,zrange=None,mask=None):
    """return the rms of the selected points."""

    points=crop_points(points,xrange=xrange,yrange=yrange,zrange=zrange,mask=mask)
    z=points[:,2]
    return np.sqrt(np.nanmean(z**2))

def histostats(points,bins=100,log=True,*args,**kwargs):
    """Plot histogram of z. Accept arguments for plt.hist.   """
    thr1=[-np.std(level_points(points)[:,2]),np.std(level_points(points)[:,2])]
    thr2=[-np.std(points[:,2]),np.std(points[:,2])]
    print('avg: %f dev:%f dev_lev:%f'%(np.mean(points[:,2]),thr1[1],thr2[1]))
    #%%timeit   #1 loops, best of 3: 19.1 s per loop
    #plt.figure(1)
    #plt.clf()
    #result=plt.hist(level_points(points)[:,2],bins=100,log=1)
    result=plt.hist(points[:,2],bins=bins,log=log,*args,**kwargs)
    #plot_points(points,shape=imsize,aspect='equal')
    plt.axvline(thr1[0],color='r')
    plt.axvline(thr1[1],color='r')
    plt.axvline(thr2[0],linestyle='-.',color='green')
    plt.axvline(thr2[1],linestyle='-.',color='green')
    #plt.title(plotTitle)
    #display(plt.gcf())
    return result

## I/O
def matrix_to_points2(mdata,x=None,y=None,xrange=None,yrange=None):
    ny,nx= mdata.shape  #changed 2015/11/04  to read ryan's data, this is because of python array shape (y,x) and because data are assumed to be in same order as image. Not sure gwyddion or nanovea data are in same format (but it should still work and giving a transposed result).
    #nx,ny= mdata.shape
    if x is None and y is None:
        if (xrange==None or yrange==None):
            #raise Exception
            x=np.arange(nx)
            y=np.arange(ny)
        else:
            x=np.linspace(*xrange,num=nx)
            y=np.linspace(*yrange,num=ny)

    xpoints,ypoints=[xy.flatten() for xy in np.array(np.meshgrid(x,y))]
    zpoints=mdata.flatten()
    points=np.vstack([xpoints,ypoints,zpoints]).T
    return points


def matrix_to_points(data,xgrid,ygrid,transpose=False):
    """this assumes that the order in data (after flattening) follows the order of
    x and y in meshgrid. If not (vertical direction first), set flag transpose.
    Not tested on arrays with different x and y, potentially a mess."""

    x,y=np.meshgrid(xgrid,ygrid)
    if transpose:
        data=data.T
    return np.vstack([x.flatten(),y.flatten(),data.T.flatten()]).T


def get_points(filename,x=None,y=None,xrange=None,yrange=None,matrix=False,addaxis=False,scale=None,center=None,skip_header=None,delimiter=','):
    """
    Return a set of xyz points (N,3) from generic csv files in xyz or matrix format.

    For example for nanovea saved txt (matrix=False) or gwyddion saved matrix (matrix=True, xrange, yrange must be defined).

    `x` and `y` added 2018/02/17 make sense only for data in matrix format,
    allowing to ovveride or set values for the x and y axis and apply
    modifications that would not be straightforward on the returned value in
    `points` format.

    input x and y can be None or empty object (calculate from data size), M-element array (must fit data size) or range (2-el). If the last two cases (when appropriate axis are passed), they are always used and get priority on what is read from file, be careful to omit them unless you want to alter data.

    This way I can e.g. open a matrix file and return points with a modified x y grid. Setting x or y to an empty object rather than to None discards axis from file and use grid indices.

    addaxis: (if matrix is set) can be set to True to read values for axis in first row and column
        (e.g. if points were saved with default addaxis=True in save_data.
    A single axis can be extracted if addaxis is set to string 'x', 'y'.
    `True` is equivalent to set to 'xy' (added 2020/06/18). 

    A complete description of the possible options is:
     read and use from file: addaxis=True, x= None
     read and discard from file, use passed: addaxis=True, x=np.array
     read and discard from file, use calculated: addaxis=True, x=[]
     don't read from file, use passed: addaxis=False, x=np.array
     don't read from file, use calculated: addaxis=False, x=None|[]

    center is the position of the center of the image in final coordinates (changed on 2016/08/10, it was '(before any scaling or rotation) in absolute coordinates.') If None coordinates are left unchanged.
        Set to (0,0) to center the coordinate system to the data.
    addaxis (if matrix is set) can be set to read values for axis in first row and column
        (e.g. if points were saved with default addaxis=True in save_points.
    2020/06/18 changes to logic in reading axis with matrix option, it was failing when x and y not set. Added possibility of individual `addAxis`
    TODO: reduce overlapping with `data2D.data_from_txt`.
    2018/02/17 reintroduced xrange even if discorauged. implemented x and y (unused as well) to axis, range
    or indices.

    """
    #import pdb
    #pdb.set_trace()
    if xrange is not None or yrange is not None:
        if verbose: 
            pdb.set_trace()
        print("WARNING: xrange and yrange options to get_points are obsolete, use x and y with two"+
               "elements to obtain same effect. Will be removed. I will correct for now")
        if x is None:
            x=xrange
        else:
            print("WARNING: xrange ignored, using X.")
        if y is None:
            y=yrange
        else:
            print("WARNING: yrange ignored, using Y.")

    #2014/04/29 added x and y as preferred arguments to xrange and yrange (to be removed).
    if skip_header is None:
        skip=0
    else:
        skip=skip_header

    mdata=np.genfromtxt(filename,skip_header=skip,delimiter=delimiter)
    if (matrix):
        if addaxis == True: addaxis = 'xy'
        
        xx = None
        yy = None
        if addaxis == 'xy':
            yy,mdata=np.hsplit(mdata,[1])
            yy=yy[1:] #discard first corner value
            xx,mdata=np.vsplit(mdata,[1])
        elif addaxis == 'x':
            xx,mdata=np.vsplit(mdata,[1])
        elif addaxis == 'y':
            yy,mdata=np.hsplit(mdata,[1])
        elif addaxis != False:
            raise ValueError("Unrecognized value for addaxis (accepted values: `x`, `y`, `xy` or boolean. Passed value: "+str(addaxis))

        # x and y can be None or empty object(calculate from data size), M-element array (must fit data size)
        #  or range (2-el). Transform them in M-el vectors
        if x is None:
            # x override xx (first line, with addAxis) if both defined
            if xx is not None:
                x = xx
            else:
                x = np.arange(mdata.shape[1])
        else:
            if np.size(x) != mdata.shape[1]:
                if np.size(x) == 2:
                    x=np.linspace(x[0],x[1],mdata.shape[1])
                else:
                    raise ValueError("X is not correct size (or 2-el array)")            
        if y is None:
            # y override yy (first col, with addAxis) if both defined
            if yy is not None:
                y = yy
            else:
                y = np.arange(mdata.shape[0])
        else:
            if np.size(y) != mdata.shape[0]:
                if np.size(x) == 2:
                    y=np.linspace(y[0],y[1],mdata.shape[0])
                else:
                    raise ValueError("Y is not correct size (or 2-el array)")            
        points=matrix_to_points2(mdata,x,y)
    else:
        points= mdata

    #2016/08/10, moved scale before center because it was setting center then scaling also the center (e.g. changing sign). With this change, center is the final center.
    if scale is not None:
        if len(scale)==2:
            scale=np.append(scale,(1.0))
        points=points*scale

    if center is not None:
        if len(center)==2:
            center=np.append(center,(0.0))
        offset=np.hstack([((np.nanmax(points,axis=0)+np.nanmin(points,axis=0))/2)[0:2],0])-np.array(center)
        points=points-offset  #center on 0


    return points

def save_points(filename,points,xgrid=None,ygrid=None,shape=None,matrix=False,fill_value=np.nan,addaxis=True,**kwargs):
    """save points on a file. If matrix is true write in matrix form (in this case you have to
    provide the values for axis). Otherwise write as points in columns."""
    #2016/10/21 rewritten routine to use points_find_grid
    #2014/08/08 default fill_value modified to nan.
    #20140420 moved filename to second argument for consistency.

    #changed interface, set automatic correction
    if type(points)==str:
        import time
        if verbose: 
            pdb.set_trace()
        print("""WARNING: routine was modified to have filename as first argument, modify IMMEDIATELY
            the calling code. Corrected automatically for this time, but I will punish you waiting 5 seconds.""")
        time.sleep(5)
        filename,points=points,filename

    if os.path.splitext(filename)[-1]=='.fits' or matrix==True:
        #fix x and y according to input
        #calculate x and y from shape if provided, then overwrite if other options
        if shape is not None:
            assert (xgrid is None) and (ygrid is None)
            s=span(points,axis=0)
            xgrid=np.linspace(s[0,0],s[0,1],shape[0])
            ygrid=np.linspace(s[1,0],s[1,1],shape[1])
        #x and y are automatically calculated, then overwrite if xgrid and ygrid provided
        x,y=points_find_grid(points,result='grid')[1] #but I don't know how to add this to fits
        if xgrid is not None:
            x=xgrid
        if ygrid is not None:
            y=ygrid
        #resample points on a matrix

        data=resample_grid(points,matrix=True,xgrid=x,ygrid=y)
        #save output
        if os.path.splitext(filename)[-1]=='.fits':
            print("creating fits..")
            hdu=fits.PrimaryHDU(data)
            hdu.writeto(filename,overwrite=1,**kwargs)  #clobber=1,**kwargs)
            return
        else:
            if addaxis:
                #add first column and row with x and y coordinates, unless flag noaxis
                #is selected.
                data=np.vstack([x,data])
                data=np.hstack([np.concatenate([[np.nan],y])[:,None],data])
                points=data
            #if not, they are already in the correct format

    points[np.isnan(points[:,2]),2]=fill_value
    np.savetxt(filename,points,**kwargs)

'''
def save_points(points,filename,xgrid=None,ygrid=None,shape=None,matrix=False,fill_value=np.nan,addaxis=True,**kwargs):
    """save points on a file. If matrix is true write in matrix form (in this case you have to
    provide the values for axis). Otherwise write as points in columns."""
    #2014/08/08 default fill_value modified to nan.
    #20140420 moved filename to second argument for consistency.
    if shape is None:
        shape=points_find_grid(points)[1]
    if matrix:
        if shape==None:
            assert len(xgrid.shape)==len(ygrid.shape)==1
        else:
            assert xgrid==ygrid==None
            x,y,z=np.hsplit(points,3)
            xgrid=np.linspace(x.min(),x.max(),shape[0])
            ygrid=np.linspace(y.min(),y.max(),shape[1])
        assert xgrid!=None
        assert ygrid!=None
        grid=np.vstack([g.flatten() for g in  np.meshgrid(xgrid,ygrid)]).T
        points=ip.griddata(points[:,0:2],points[:,2],grid,fill_value=fill_value,method=method)
        points=points.reshape(ygrid.size,xgrid.size)
        if addaxis:
            #add first column and row with x and y coordinates, unless flag noaxis
            #is selected.
            points=np.vstack([xgrid,points])
            points=np.hstack([np.concatenate([[np.nan],ygrid])[:,None],points])
        #points=np.hstack([grid,points[:,np.newaxis]])
    #if not, they are already in the correct format
    if os.path.splitext(filename)[-1]=='.fits':
        if matrix:
            print "creating fits.."
            hdu=fits.PrimaryHDU(points)
            hdu.writeto(filename,clobber=1,**kwargs)
            return
        else:
            print 'Fits can be saved only for matrix data.'
    else:
        np.savetxt(filename,points,**kwargs)

'''

#SHAPE AND RESAMPLING FUNCTIONS

def points_find_grid2(points,result='shape',sort=None,steps=None):
    """Given points as pointcloud, do some basic guess on shape and axis orientation
    of the grid.
    Works for raster points, even irregular and non rectangular, but not for scatter
    (step is estimated from the first two elements, it fails if elements are not sorted,
    points can be sorted xy or yx to avoid failure, like (xysort):
        a = a[a[:,1].argsort()]
        a = a[a[:,0].argsort(kind='mergesort')] #stable sort.
    Return a tuple (fastind,result), fastind is 1 for y and 0 for x,
    result can be 'shape', 'step'  or 'grid'.
    step can be provided as scalar or 2d vector to enforce step size

    sort can be None, 'xy', 'yx' (respectively try to guess, x or y faster)
        or 'none' (apply stable sort xy before calculating grid).

    This second version, works well in well-behaved cases, for more complex situations,
    better to check results (see critical cases below).

    """

    """N.B.: it fails on unsorted points:
    ff=r'C:\\Users\Vincenzo\Google Drive\Shared by Vincenzo\Metrology logs and data\Simulations\Coating_stress\MirrorStressIF\1mOAB_100x100\Mirror_Stress_IF_Iridium_C_total.csv'
    pf=get_points(ff,delimiter=' ')
    points_find_grid(pf)[1]
    gives (10498, 100)

    it fails on some data sets, e.g. PCO1.2S02_BCB_rep2 in POOL\pipeline\surfaceNB2
    """

    #2018/02/12 removed failed attempt of sorting points before result
    #in points_find_grid2
    #2018/09/29 readded at beginning according to docstring example

    # determine fastindex
    if sort == 'xy':
        fastind=0
    elif sort == 'yx':
        fastind=1
    elif sort is None:
        #try to guess
        #determines the fastest from the smallest step.
        # if steps are the same cannot determine. Use some point at the middle
        # rather than first point to reduce chance of error
        i = np.array([-1,0,1])+int(points.shape[0]/2)  #central index
        #three values of i are set to avoid it happens at at the end of a line
        d=[points[i+1,0]-points[i,0],points[i+1,1]-points[i,1]] #distances x and y from next point
        #d=[points[1,0]-points[0,0],points[1,1]-points[0,1]]
        if np.abs(min(d[0]))<np.abs(min(d[1])):
            fastind=0  #y in points
        elif np.abs(min(d[0]))>np.abs(min(d[1])):
            fastind=1   #x
        else:
            fastind=np.nan
    elif sort == 'none':
        #sort them as xy
        fastind=0
        points = points[points[:,1].argsort()]
        points = points[points[:,0].argsort(kind='mergesort')] #stable sort.
    else:
        raise ValueError ('sort must be "xy", "yx" or None, instead of %s '%sort)

    #determine nfast as the length of first change of sign in
    # calculate the shape of the matrix detecting
    # first change of sign of derivative of scanning
    # coordinate (end of line and return to first point)
    # in fast axis.
    sign=np.sign(d[fastind])
    #nfast=(np.sign(np.diff(points[:,fastind]))!=sign)[0]).sum()+1
    id=np.where(np.sign(np.diff(points[:,fastind]))!= sign)[0] #positions of change of sign, it's last point in line
    id=np.hstack([id,[points.shape[0]-1]])
    nslow=np.max(np.diff(id)) #diff is the length of each line
    nfast=len(id)
    #nslow=points.shape[0]/nfast
    if nslow*nfast!=points.shape[0]:
        if verbose: 
            pdb.set_trace()
        print("""WARNING: points number [%i] doesn't match regular grid for size determined by points_find_grid [%i x %i],
        usually OK, but please double check results."""%(points.shape[0],nslow,nfast))

    if steps is not None:
        if steps==(0,0):
            steps=d #automatically determine
        raise NotImplementedError

    slowind=int(not(fastind))
    if result=='step':  #uses d
        if steps is not None:
            retval = steps
        else:
            steps=[d[fastind],d[fastind]]
            slowsteps=np.diff(points[id,slowind])
            steps[slowind]=slowsteps[np.argmax(np.abs(slowsteps))]
            retval=steps
    elif result=='shape':  #uses nslow
        if steps is not None:
            retval = steps
        else:
            #here axis are switched meaning the position of index to
            # access the matrix, i.e. the matrix has shape (ny,nx)
            shape=[nslow,nslow]
            shape[slowind]=nfast
            retval=shape
    elif result=='grid':  #use points to calculate by points_find_grid
        #xs,ys,zs=np.hsplit(span(points,axis=0),3)
        xs,ys,zs=span(points,axis=0)
        nx,ny=points_find_grid(points)[1]
        xgrid=np.linspace(xs[0],xs[1],nx)
        ygrid=np.linspace(ys[0],ys[1],ny)
        retval=[xgrid,ygrid]
    else:
        raise ValueError

    return fastind,tuple(retval)

def points_find_grid(points,result='shape',sort=None):
    """Given points as pointcloud, do some basic guess on shape and axis orientation
    of the grid. Not many checks in this first version, you need to verify results.
    Works for raster points, even irregular and non rectangular, but not for scatter (step is estimated from the first two
    elements, it fails if elements are not sorted, points can be
    sorted xy or yx to avoid failure, like (xysort):
        a = a[a[:,1].argsort()]
        a = a[a[:,0].argsort(kind='mergesort')] #stable sort.
    Return a tuple (fastind,result), fastind is 1 for y and 0 for x,
    result can be 'shape', 'step'  or 'grid'."""

    """N.B.: it fails on unsorted points:
    ff=r'C:\\Users\Vincenzo\\Google Drive\Shared by Vincenzo\Metrology logs and data\Simulations\Coating_stress\MirrorStressIF\1mOAB_100x100\Mirror_Stress_IF_Iridium_C_total.csv'
    pf=get_points(ff,delimiter=' ')
    points_find_grid(pf)[1]
    gives (10498, 100)

    it fails on some data sets, e.g. PCO1.2S02_BCB_rep2 in POOL\pipeline\surfaceNB2"""

    #2018/02/12 removed failed attempt of sorting points before result
    #in points_find_grid2

    #determines the fastest from the smallest step.
    # if steps are the same cannot determine.
    d=[points[1,0]-points[0,0],points[1,1]-points[0,1]]
    if np.abs(d[0])<np.abs(d[1]):
        fastind=1  #y in points
    elif np.abs(d[0])>np.abs(d[1]):
        fastind=0   #x
    else:
        fastind=np.nan

    # calculate the shape of the matrix detecting
    # first change of sign of derivative oof scanning
    # coordinate (end of line and return to first point)
    # in fast axis.
    sign=np.sign(d[fastind])
    #nfast=(np.sign(np.diff(points[:,fastind]))!=sign)[0]).sum()+1
    id=np.where(np.sign(np.diff(points[:,fastind]))!= sign)[0] #positions of change of sign, it's last point in line
    id=np.hstack([id,[points.shape[0]-1]])
    nslow=np.max(np.diff(id)) #diff is the length of each line
    nfast=len(id)
    #nslow=points.shape[0]/nfast
    if nslow*nfast!=points.shape[0]:
        if verbose: 
            pdb.set_trace()
        print("""WARNING: points number [%i] doesn't match regular grid for size determined by points_find_grid [%i x %i],
        usually OK, but please double check results."""%(points.shape[0],nslow,nfast))

    slowind=int(not(fastind))
    if result=='step':
        steps=[d[fastind],d[fastind]]
        slowsteps=np.diff(points[id,slowind])
        steps[slowind]=slowsteps[np.argmax(np.abs(slowsteps))]
        retval=steps
    elif result=='shape':
        #here axis are switched meaning the position of index to
        # access the matrix, i.e. the matrix has shape (ny,nx)
        shape=[nslow,nslow]
        shape[slowind]=nfast
        retval=shape
    elif result=='grid':
        #xs,ys,zs=np.hsplit(span(points,axis=0),3)
        xs,ys,zs=span(points,axis=0)
        nx,ny=points_find_grid(points)[1]
        xgrid=np.linspace(xs[0],xs[1],nx)
        ygrid=np.linspace(ys[0],ys[1],ny)
        retval=[xgrid,ygrid]
    else:
        raise ValueError

    return fastind,tuple(retval)

def rebin_points(tpoints,matrix=False,steps=None,*args,**kwargs):
    """Use `stats.binned_statistic_2d` to create a new surface from original points
    divided in bins centered on xy of final points. Points in a bin can be averaged
    (default) or binned statistics options can be used to obtain result values.
    
    It is theoretically the best way to create a surface on a grid with minimal loss of information,
    however it can be tricky, both for estimating errors (or values on empty bins) and for 
    transformations (e.g. crop )."""
    
    
    x,y,z=list(zip(*tpoints))
    #T=kwargs.pop('transpose',False)
    xr=span(x)
    yr=span(y)

    if steps is not None:
        try:    #2-el
            if len(steps)==1:
                steps=[steps,steps]
            elif len(steps)!=2:
                raise ValueError ("1 or 2d only accepted.")
        except:  #1-el
            steps=[steps,steps]

        #this is necessary to keep step constant,
        #  note that #of points (inclusion of end) can vary according
        #  to floating point behavior of arange.
        xb=np.arange(*xr,step=steps[0])-steps[0]/2.
        xb=np.append(xb,xb[-1]+steps[0])
        yb=np.arange(*yr,step=steps[0])-steps[1]/2.
        yb=np.append(yb,yb[-1]+steps[1])
        bins=[xb,yb]

    else:
        bins=kwargs.pop('bins',None)

    ss=stats.binned_statistic_2d(x,y,z,bins=bins,*args,**kwargs)
    x2=np.array([(xx+yy)/2. for xx,yy in zip(ss[1][:-1],ss[1][1:])])
    y2=np.array([(xx+yy)/2. for xx,yy in zip(ss[2][:-1],ss[2][1:])])

    if matrix:
        return ss[0].T,x2,y2
    else:
        return matrix_to_points(ss[0],x2,y2)

"""
d=np.vstack([f.flatten() for f in np.meshgrid(arange(3),arange(3)+10.)]).T
d=vstack([d.T,d[:,0]**2+d[:,1]**2]).T
"""

## resample_points: resample points on other points and return points
## resample_grid: resample points on square grid. Return points or matrix (no x and y that is a bit confusing since they are optional argument)
## points_autoresample:  return data2D in form data,x,y automatically calculated

## crop problem: if I crop points and then do auto_resample, in general I can get a frame of nan.
##      Think to points not on a grid. When I crop the result is still not on a grid, so when I resample
##      edge points are lost. Possible solution is to interpolate first on a regular grid, then crop.
##      This is ok, but requires understanding from the user and most likely requires to use crop_data
##      in data2D. The argument cut was also introduced in resample_grid to overcome this problem
##      in a brutal way. Another possibility is to introduce crop interval as argument.

def resample_points(tpoints,positions):
    """resample tpoints [Npoints x 3] on the points defined in positions [Mpoints x 2], or [Mpoints x 3]
    (in this case 3rd column is ignored).
    Return a [Nx x Ny , 3] vector of points. To get a (plottable) matrix of data use:
    plt.imshow(rpoints[:,2].reshape(xgrid.size,ygrid.size))."""
    assert tpoints.shape[1]==3
    z=ip.griddata(tpoints[:,0:2],tpoints[:,-1],positions[:,0:2],method=method)
    rpoints=np.hstack([positions[:,0:2],z[:,np.newaxis]])
    return rpoints

def resample_grid(tpoints,xgrid=None,ygrid=None,matrix=False,resample=True):

    """resample tpoints [Npoints x 3] on the grid defined by two vectors xgrid [Nx] and ygrid [Ny].
    Return a [Nx * Ny , 3] vector of points, sorted in standard python order
    (x changes faster) or a matrix if matrix=True. if resample is set to False
    only x and y are changed and values are not touched (must maintain number of points).
    matrix=True-->points to matrix
    p=resample_grid(p) #straighten the grid of p changing data as little as possible
    matrix=False,resample=False-> Convert from matrix to points without resampling if matrix input, useless if input is points (do two opposite operations that should cancel each other).

    """
    """ old (matrix=False):
    To get a (plottable) matrix of data use:
    plt.imshow(rpoints[:,2].reshape(ygrid.size,xgrid.size))."""

    assert tpoints.shape[1]==3
    if (xgrid is None) and (ygrid is None):
        xgrid,ygrid=points_find_grid(tpoints,'grid')[1]

    tpoints=tpoints[np.isfinite(tpoints[:,2]),:]  #2018/10/01
    x,y=np.meshgrid(xgrid,ygrid)
    if resample:  #if both resample and matrix are False, two useless operations are performed and the final array is unchanged.
        z=ip.griddata(tpoints[:,0:2],tpoints[:,2],(x,y),method=method) #this is super slow, but still faster than the one in matplotlib
    else:
        z=tpoints[:,2].reshape(ygrid.size,xgrid.size)
    if matrix:
        return z
    else:
        return np.vstack([x.flatten(),y.flatten(),z.flatten()]).T
    #2015/08/31
    #xy=zip(*(x.flat for x in np.meshgrid(xgrid,ygrid)))
    #return np.vstack([np.array(xy).T,data.flatten()]).T #magical formula found by trial and error
    #2014/11/24 rpoints=np.vstack([x.T.flatten(),y.T.flatten(),z.T.flatten()]).T
    #return rpoints


def points_autoresample(points,cut=0,resample=True):
    """Use points_find_grid to determine the grid for points and resample.
    It should give minimal alteration of points when close to grid.
    This routine returns data,x,y
    Same result can be obtained by calling
    resample_grid (without providing x and y)
    AND points_find_grid to determine x and y."""

    '''
def points_autoresample(points,edge=0):
    """Edge is the number of points to cut at edges."""
    sh=points_find_grid(points)[1]
    st=points_find_grid(points,result='step')
    x=np.linspace(span(points,axis=0)[0][0],span(points,axis=0)[0][1],sh[0]) #do some cropping to cut out nan
    x=x[edge:len(x)-edge]
    y=np.linspace(span(points,axis=0)[1][0],span(points,axis=0)[1][1],sh[1])
    y=y[edge:len(y)-edge]
    '''
    #print 'obsolete, replace with resample_grid(points,matrix=True)'
    fastax,(x,y)=points_find_grid(points,result='grid')
    if cut>0:
        x=x[cut:-cut]
        y=y[cut:-cut]
    data=resample_grid(points,x,y,matrix=True,resample=resample)
    return data,x,y



def extract_profile(points,xy0,xy1=None,npoints=None,along=True,plot=False):
    """extract a profile from xy0=(x0, y0) to xy1=(x1,y1).
    Return a couple of vectors x, y, z. The number of points can be set, otherwise is set
    accordingly to the longest profile dimension.
    If `along` is set (default), a two-dim x-z profile is returned with x distance
    along the profile from xy0.
    If called without second argument, extract point.
    #TODO, can be made faster by cropping data around profile.
    """
    
    # npoints not provided uses coordinates
    ## infer grid
    _,(xg,yg) = points_find_grid(points,result='grid') #complete grid in xg,yg

    if xy1 is None: #extract point
        if len(np.shape(xy0)) == 1:
            xy1 = xy0
        else:
            #iterate
            profiles = [extract_profile(points,p[0],p[1],npoints=npoints,
                along=along,plot=(i==0 & plot)) for i,p in enumerate(xy0)]
            if plot :
                ax1 = plt.gcf().axes[0]
                ax2 = plt.gcf().axes[-1]
                for p,prof in zip(xy0[1:],profiles[1:]):
                    #ax1 = plt.subplot(211,label='surf')
                    plt.sca(ax1)
                    p0,p1=p
                    plt.plot([p0[0],p1[0]],[p0[1],p1[1]])
                    #ax2 = plt.subplot(212,label='profs') #,sharex=ax1)       
                    plt.sca(ax2)
                    #pdb.set_trace()
                    r= prof[0]
                    z= prof[-1]
                    plt.plot(r,z)
            return profiles
            
    #span of extreme points
    xs = np.sort([xy0[0],xy1[0]])  
    ys = np.sort([xy0[1],xy1[1]])
    
    #  range for crop of useful data,
    ix = np.where((xg>=xs[0]) & (xg<=xs[1]))[0]
    iy = np.where((yg>=ys[0]) & (yg<=ys[1]))[0]
    #  add one points to each side
    if len(ix) == 0:
        ix=[(np.abs(xg - np.nanmean(xs))).argmin()]
    if ix[0] !=0: ix = np.insert(ix,0,ix[0]-1)
    if ix[-1] !=(len(xg)-1): ix = np.append(ix,ix[-1]+1)
    if len(iy) == 0:
        iy=[(np.abs(yg - np.nanmean(ys))).argmin()]
    if iy[0] !=0: iy = np.insert(iy,0,iy[0]-1)
    if iy[-1] !=(len(yg)-1): iy = np.append(iy,iy[-1]+1)
    
    if npoints is None:
        npoints=max(len(ix),len(iy))
    else:
        if xy0 == xy1:
            raise ValueError("npoints cannot be provided for single point")
    
    #adjusted range
    xr = span(xg[ix])
    yr = span(yg[iy])
    
    pp = crop_points(points,xr,yr)
    #points = crop_points(points,[xy0[0],xy1[0]],[xy0[1],xy1[1]])
    
    if np.all(xy0 == xy1):
        return ip.griddata(pp[:,0:2],pp[:,-1],xy0,method=method)
    
    xx=np.linspace(xy0[0],xy1[0],npoints)
    yy=np.linspace(xy0[1],xy1[1],npoints)
    r=np.sqrt((xx-xy0[0])**2+(yy-xy0[1])**2)

    z=ip.griddata(pp[:,0:2],pp[:,-1],np.vstack([xx,yy]).T,method=method)
    #pdb.set_trace()
    if along:
        #pp=np.vstack([r,z]).T
        pp=[r,z]
    else:      #pp=np.hstack([xx[:,np.newaxis],yy[:,np.newaxis],z[:,np.newaxis]])
        pp=[xx,yy,z]
        r=xx

    if plot:
        #plt.figure()
        ax1 = plt.subplot(211)
        plot_points(points,aspect='auto')
        plt.plot([xy0[0],xy1[0]],[xy0[1],xy1[1]],'r')
        xl=plt.xlim()
        ax2 = plt.subplot(212) #,sharex=ax1)
        if not along:
            ax1.get_shared_x_axes().join(ax1, ax2)
        plt.plot(r,z)
        #plt.xlim(*xl)
        plt.grid()
        #plt.colorbar().remove() #dirty trick to adjust size to other panels
        #this works better
        #pdb.set_trace()
        box1 = ax1.get_position()
        box2 = ax2.get_position()
        ax2.set_position([box1.x0, box2.y0, box1.width , box2.height])
        #pdb.set_trace()
    
    return pp #points[:,0],points[:,1]

'''
def plot_markers(m,subplots=0,points=None,w=None,**kwargs):
    """m is N points x,y {Nx2}. plot circles around points.
    if subplots is set, generate subplots with zoomed area."""

    if subplots and (points is not None):
        nrows = int(np.sqrt(len(m[:,0])))+1 #int(math.ceil(len(subsl) / 2.))
        plt.figure()
        fig, axs = plt.subplots(nrows, nrows,squeeze=True)
        for marker,ax in zip(m,axs.flatten()[0:len(m)]):
            plt.sca(ax)
            scatter=scatter if kwargs.has_key('scatter') else True
            plt.xlim(marker[0:1]+[-w/2,w/2])
            plt.ylim(marker[1:2]+[-w/2,w/2])
            plot_points(points,scatter=scatter,bar=0,**kwargs)
            # Make an axis for the colorbar on the right side
        for ax in axs.flatten()[len(m):]:
            plt.cla()
        cax = fig.add_axes([0.9, 0.1, 0.03, 0.8])
        plt.colorbar()
    else:
        plt.plot(m[:,0],m[:,1],'o',markerfacecolor='none', c='w',lw=20,markersize=20)
'''

def plot_points(points,xgrid=None,ygrid=None,shape=None,units=None,resample=True,scatter=False,contours=0,bar=True,stats=True,
    **kwargs):
    #should be named show
    """resample xyz points [Nx3] to a grid whose axis xgrid and ygrid are given
    and plot it. If resample is set to False x and y positions are considered only for range,
    but they are not used to position the z values (it works if x and y are on an exact unrotated grid,
    resampling is slower, but exact).
    shape is in format (nx,ny) that is opposite to python convention.
    contours is contour spacing."""

    from pySurf.data2D import plot_data #here to avoid circular import


    import pdb
    #plot
    #plt.ioff()
    if np.size(units)==1:
        units=np.repeat(units,3)

    x,y,z=np.hsplit(points,3)
    #cmap=kwargs.pop('cmap','jet')
    aspect=kwargs.pop('aspect','equal') #needed because scatter plot don't have kw aspect
    #print id(points)
    #pdb.set_trace()
    plt.gca().set_aspect(aspect)
    if contours:
        scatter=False
        bar=False

    if scatter: #do scatterplot
        beamsize=kwargs.pop('s', 20) #I may use this to represent beamsize (round symbol) or lateral resolution (square)
        plt.scatter(x, y, c=z, s=beamsize, edgecolors='None', **kwargs)
        #plt.xlim(span(x))
        #plt.ylim(span(y))
    else:
        #if not scatter plot, grid can be provided as shape or as xgrid and ygrid axis. If not provided is automatically guessed.

        if xgrid is None and ygrid is None:
            if shape is None:
                shape=points_find_grid(points)[1]
            #xgrid=np.linspace(x.min(),x.max(),shape[0])
            #ygrid=np.linspace(y.min(),y.max(),shape[1])
            xgrid,ygrid=points_find_grid(points,'grid')[1]  #2018/10/10
        else:
            if not len(xgrid.shape)==len(ygrid.shape)==1:
                raise ValueError

        if resample:
            print("resampling...")
            z=resample_grid(points,xgrid,ygrid)[:,2]

        ## consider replacing all of the following with call to plot_data
        nx,ny=shape  #[xx.size for xx in (xgrid,ygrid)]
        xr,yr,zr=[span(xx) for xx in (xgrid,ygrid,z)]  #ranges
        #xxg,sx=np.linspace(xr[0],xr[1],nx,retstep=1)   #grids (not needed anywhere) and steps (needed only to adjust plot on pixel centers.
        #yyg,sy=np.linspace(yr[0],yr[1],ny,retstep=1)
        ##ranges for plot (intervals centered on xxg, yyg)
        #xr=xr+np.array((-sx/2,sx/2))
        #yr=yr+np.array((-sy/2,sy/2))

        xr = span_from_pixels(xr,nx)
        yr = span_from_pixels(yr,ny)

        z=z.reshape(ny,nx)

        if contours:
            levels=np.arange(zr[0],zr[1],contours)
            #CS=plt.contour(xgrid,ygrid,z,colors='b',extent=[xr[0],xr[1],yr[0],yr[1]],aspect=aspect,levels=levels,
            #    origin='lower', **kwargs)
            #version without extent
            CS=plt.contour(xgrid,ygrid,z,colors='b',aspect=aspect,levels=levels, \
                    origin='lower', **kwargs)
        else:
            plot_data(z,xgrid,ygrid, **kwargs)
            bar=False
            #plt.imshow(z,extent=[xr[0],xr[1],yr[0],yr[1]],interpolation='none',aspect=aspect,
            #origin='lower', **kwargs)
    
    plt.xlabel('X'+(" ("+units[0]+")" if units[0] is not None else ""))
    plt.ylabel('Y'+(" ("+units[1]+")" if units[1] is not None else ""))
    
    if stats:
        from plotting.captions import legendbox
        from pySurf.data2D import get_stats
        if scatter:
            xgrid,ygrid=x,y
        legendbox(get_stats(z,xgrid,ygrid))
    
    if bar:
        cb=plt.colorbar()
        if units[2] is not None:
            cb.ax.set_title(units[2])
    
    #plt.gca().autoscale(False) #2021/03/29, not sure what this was for, 
    #   it was giving wrong plot with scatter = True, removed.
    #plt.show()
    #plt.ion()
    
    return z
    

def subtract_points(p1,p2,xysecond=False,resample=True):
    """Subtract second set of points after interpolation on first set coordinates.
    If xySecond is set to True data are interpolate on xy of p2 and then subtracted."""

    #2018/02/15 moved later to save one copy in default case
    #p1=p1.copy()  #added 2015/01/13 together with resample option.
    #p2=p2.copy()  #this is unnecessary

    if resample:
        if not(xysecond):
            p2=resample_points(p2,p1)
        else:
            p1=resample_points(p1,p2)
            p2=p2.copy()
    else:
        p2=p2.copy()
    #at this point p1 and p2 are on same grid

    p2[:,2]=p1[:,2]-p2[:,2] #note that this is a slice, so array is modifyied

    return p2

def subtract_points2(p1,p2,xySecond=False,resample=True):
    """attempt to make subtract_points faster by transposing and acting on faster slice"""

    if resample:
        if not(xySecond):
            p2=resample_points(p2,p1)
        else:
            p1=resample_points(p1,p2)
            p2=p2.copy()
    else:
        p2=p2.copy()
    #at this point p1 and p2 are on same grid

    p2.T[2,:]=p1.T[2,:]-p2.T[2,:] #note that this is a slice, so array is modifyied

    return p2

def subtract_points_old(p1,p2,xySecond=False,resample=True):
    """Subtract second set of points after interpolation on first set coordinates.
    If xySecond is set to True data are interpolate on xy of p2 and then subtracted."""

    p1=p1.copy()  #added 2015/01/13 together with resample option.
    p2=p2.copy()  #this is unnecessary

    if resample:
        if not(xySecond):
            p2=resample_points(p2,p1)
        else:
            p1=resample_points(p1,p2)

    p1[:,2]=p1[:,2]-p2[:,2] #note that this is a slice, so array is modifyied

    return p1


#FUNCTIONS THAT CAN BE DONE WITH MASKS

def roicircle_points(points,radius,missing=None,invert=False):
    """Select points inside a circle of given radius. Points outside the circle are removed
    or replaced by a missing value."""
    #this is downsampling points and defining pts.

    aa=(points[:,0]**2+points[:,1]**2)<(radius**2)
    if invert:
       aa=np.logical_not(aa)
    na=np.logical_not(aa)   #negative of aa
    if missing is None:
        points=points[aa,:]
    else:
        points[na,2]=missing

    return points

"""
def clip_points(min=min,max=max,clipvalue=clipvalue,nsigma=nsigma,
        refimage=refimage,mask=mask):
    #TODO accept points as refimage
    #TODO if ADDMARKER is set a marker is added for clipped points

    clippeddata=clip( data, min=min,max=max,clipvalue=clipvalue,nsigma=nsigma,$
      refimage=refimage,mask=mask,torefimage=torefimage,$
      _strict_extra=extra)
"""

##DEVELOPMENT


def smooth_points(points,xywidth,xgrid=None,ygrid=None,shape=None,matrix=False):
    """resample points on a grid and perform moving average smoothin in x and y according to xywidth,
    if one component is None smoothing in that direction is not performed.
    Return points in usual coordinates, unless matrix flag is set."""
    data=points_to_matrix(points,xgrid,ygrid)
    if not(xywidth[0] is None):
        data=np.apply_along_axis(running_mean,0,data,xywidth[0])
    if not(xywidth[1] is None):
        data=np.apply_along_axis(running_mean,1,data,xywidth[1])
    if matrix:
        return data
    else:
        return matrix_to_points(data,xgrid,ygrid)


def smooth_points2(points,xywidth,xgrid=None,ygrid=None,shape=None):
    """NOT WORKING shape in format (nx,ny) as convention in points, opposite of python."""
    def gauss_kern(size, sizey=None):
        """ Returns a normalized 2D gauss kernel array for convolutions """
        size = int(size)
        if not sizey:
            sizey = size
        else:
            sizey = int(sizey)
        x, y = mgrid[-size:size+1, -sizey:sizey+1]
        g = exp(-(x**2/float(size)+y**2/float(sizey)))
        return g / g.sum()

    def blur_image(im, n, ny=None) :
        """ blurs the image by convolving with a gaussian kernel of typical
            size n. The optional keyword argument ny allows for a different
            size in the y direction.
        """
        g = gauss_kern(n, sizey=ny)
        improc = signal.convolve(im,g, mode='valid')
        return(improc)

    x,y,z=np.hsplit(points,3)
    if shape==None:
        if not len(xgrid.shape)==len(ygrid.shape)==1:
            return z    #skip plot
    else:
        assert xgrid==ygrid==None
        xgrid=np.linspace(x.min(),x.max(),shape[0])
        ygrid=np.linspace(y.min(),y.max(),shape[1])
    z=resample_grid(points,xgrid,ygrid)[:,2]
    nx,ny=[xx.size for xx in (xgrid,ygrid)]
    xr,yr=[span(xx) for xx in (xgrid,ygrid)]
    xxg,sx=np.linspace(xr[0],xr[1],nx,retstep=1)
    yyg,sy=np.linspace(yr[0],yr[1],ny,retstep=1)
    #ranges for plot (intervals centered on xxg, yyg)
    xr=xr+np.array((-sx/2,sx/2))
    yr=yr+np.array((-sy/2,sy/2))
    z=z.reshape(ny,nx)

    xwidth,ywidth=xywidth

    z=blur_image(z,xwidth,ywidth)
    plt.imshow(z,extent=[ygrid.min(),ygrid.max(),xgrid.max(),xgrid.min()],aspect=aspect,interpolation='none')
    plt.xlabel('X (mm)')
    plt.ylabel('Y (mm)')
    plt.colorbar()
    #plt.show()
    return z

if __name__=="__main__":
  pts=get_points(gfile, delimiter='')
  
  


import numpy as np
import math
from .points import rotate_points,translate_points
import logging
import itertools

from matplotlib import pyplot as plt

## rotate_center -> points.rotate_points
## translate -> points.translate_points

def rot_center_func(theta,center=(0,0)):
    def func(x):
        return rotate_points(x,theta,center)
    return func
    
def translate_func(offset=(0,0)):
    def func(x):
        return translate(x,offset)
    return func
    
def rototrans_func(theta,center=(0,0),offset=(0,0)):
    """return a function that rotate by theta about center THEN translate by offset"""
    def func(x):
        if (x.shape[-1]==3):
            xy=x[:,0:2]
            return np.hstack([translate_points(rotate_points(xy,theta,center),offset),x[:,2,np.newaxis]]) 
        else:  
            return translate_points(rotate_points(x,theta,center),offset)
    return func

def apply2D(func,*args,**kwargs):
    """given a function acting on 2D points as array[N,2], returns a function
    that do same action on first two coordinates of a 3D array, leaving the third unchanged.
    e.g. rototrans_func=apply2D(translate_points(rotate_points()))
    """
    
    def func3d(x):
        if (x.shape[-1]==3):
            xy=x[:,0:2]
            return np.hstack([func(xy,*args,**kwargs),x[:,2,np.newaxis]]) 
        else:  
            return func(x,*args,**kwargs)
            
    return func3d
    
def _angpos2(xy):
    '''given some notable points, return the angular positions 
    of points with respect to barycenter. The angle returned is in the range [-pi:pi].
    Return angle and radius for each point and barycenter position.'''

    #N xy points are expressed as array N x 2
    b=(np.sum(xy,0))/(xy.shape[0])    #calculate barycenter
    deltaxy=xy-(b.repeat(xy.size/2).reshape(2,-1)).T
    r=np.sqrt(np.sum(deltaxy**2,1))
    theta=np.arctan2(deltaxy[:,1],deltaxy[:,0])
    return theta,r,b    


def plot_transform(points,plotLines=None,labels=None,transform=None):
    """plot a set of points or shapes and their transformed.
    points is a list of sets of coordinates, one for each groups of points to be plotted with same style and transformation. Stile and transformation can be passed as single value or list.
    plotlines is an array of flags to plot as lines(True) or points (False,default).
    labels set of labels of same length
    transform list (same len as points) of transformation functions or single function to be 
    applied respectively to each set of points passed.
    
    #this plots some markers and a rectangle (see examples in module):
    plt.figure() plot_transform([primary,secondary,rect1],[0,0,1],['markers1','markers2','ROI'],[transform2,None,transform2])
    plt.show()
    display(plt.gcf())
    """
    
    syms = itertools.cycle((',', '+', 'x', 'o', '*')) 

    if not(hasattr(points,'__len__')): #transform in list if passed as scalar
        points=[points]
    elif not(hasattr(transform,'__len__')):
        transform=np.repeat(transform,len(points))
    points= [np.array(p) for p in points]
    if hasattr(transform,'__len__') and len(transform)==1:
        transform=np.repeat(transform,len(points))
    elif not(hasattr(transform,'__len__')):
        transform=np.repeat(transform,len(points))
    if plotLines is None:
        plotLines=np.repeat(False,len(points))
    if labels is None:
        labels=np.repeat('_nolegend_',len(points))
        
    #plt.clf()
    plt.grid(1)
    plt.gca().set_aspect('equal')
    for markers1,pl,l,trans in zip (points,plotLines,labels,transform):
            
        if pl:
            plt.plot(markers1[:,0],markers1[:,1],label=l)
        else:
            plt.plot(markers1[:,0],markers1[:,1],next(syms),label=l)

        if trans is not None:
            m=trans(markers1)
            if pl:
                plt.plot(m[:,0],m[:,1],label=l+' trans.')
            else:
                plt.plot(m[:,0],m[:,1],next(syms),label=l+' trans.')           
    
    # Shrink current axis by 20%
    if np.any(points):
        ax=plt.gca()
        box = ax.get_position()
        ax.set_position([box.x0, box.y0, box.width * 0.8, box.height])

        # Put a legend to the right of the current axis
        ax.legend(loc='center left', bbox_to_anchor=(1, 0.5))
        #plt.legend(loc=0)
    else:
        print('no markers to plot')

def find_rototrans(markers1,markers2,pars=None):
    """Return the transformation that applied to markers1 gives markers2.
    If list is passed in pars, transformation parameters mrot,b1,bartrans are appended."""
    #used by calibrate_align as find_transform
  
    markers1=np.array(markers1)
    markers2=np.array(markers2)
    
    assert np.ndim(markers1)==2
    assert markers1.shape==markers2.shape
    
    gang0,gr,gb=_angpos2(markers1)
    mang0,mr,mb=_angpos2(markers2)
    
    #determine the rotation angle mrot.
    #Note that this is not correct, angles should be weighted by distance from barycenter:
    #   an error on x and y has differenct impact on the angular position of point according
    #   to distance from the center.
    #note that if there is a large rotation, angles can be messed up by preiodicity,
    #   need to account for that.
    mrot0=gang0[0]-mang0[0] #angular displacement between first elements
    #g-gang0[0] and g-mang0 are in the range [-2pi,2pi], transform them in range [0,2pi]
    gang=np.array([(g-gang0[0]) if (g-gang0[0])>=0 else (g-gang0[0]+2*math.pi) for g in gang0])
    mang=np.array([(g-mang0[0]) if (g-mang0[0])>=0 else (g-mang0[0]+2*math.pi) for g in mang0])
    mrot=(gang-mang).mean()+mrot0
    
    bartrans=gb-mb
    try: 
        logger
    except NameError:
        logger=logging.getLogger()
        logger.setLevel(logging.DEBUG)
    #logger=logging.getLogger()
    logger.info("stdv of markers distance errors from barycenter: %s"%((gr-mr).std()))
    logger.info('rotation angle (degrees): %s +-%s'%(mrot*180./math.pi,(gang-mang).std()*180./math.pi))
    logger.info('translation: %s'%(bartrans))

    #define the transformation function. I will rotate glass data
    trans=rototrans_func(-mrot,gb,-bartrans)  #(mrot,mb,bartrans) to rotate mandrel 
    #if gmarkers.shape[0]: transform=a2d.find_affine(mmarkers,gmarkers)
    err=markers2-trans(markers1)
    logger.info("errors in markers position after rotations: \n%s"%(err))
    logger.info("mean of error abs val: \n%s"%(np.sqrt((err**2).sum(axis=1)).mean()))
    if pars is not None:
        pars.extend([mrot,b1,bartrans])
    return trans 

def find_affine(markers1, markers2,pars=None):
    """Return a function that can transform points from the first system to the second. 
    if pars is set to a list, append the matrix A of the transformation, that can be applied to a vector x with:
    unpad(np.dot(pad(x), A))
    markers1 and markers2 are sets of points in format [Npoints, Ndim]. Transformation matrix A is [Ndim+1 x Ndim+1].
    
    http://stackoverflow.com/questions/20546182/how-to-perform-coordinates-affine-transformation-using-python-part-2
    """
    markers1=np.array(markers1)
    markers2=np.array(markers2)
    
    # Pad the data with ones, so that our transformation can do translations too
    n = markers1.shape[0]
    pad = lambda x: np.hstack([x, np.ones((x.shape[0], 1))])  #adds a column of ones to the right 
    unpad = lambda x: x[:,:-1]   #remove rightmost column
    X = pad(markers1)
    Y = pad(markers2)

    # Solve the least squares problem X * A = Y
    # to find our transformation matrix A
    A, res, rank, s = np.linalg.lstsq(X, Y)
    
    #note: the transformation to and from matrix is needed to make it work also if x is a single point.
    trans = lambda x: np.array(unpad(np.dot(pad(np.matrix(x)), A)))
    if pars is not None:
        pars.append(A)
    
    #calculate parameters and log copied from find_rototrans, ideally should be 
    #   derived from transformation matrix. see e.g.
    # https://trac.osgeo.org/postgis/wiki/DevWikiAffineParameters
    #
    # empirically:
    # A[2,0]= offset
    # A[2,1]= rotation angle
    # A[0,1], A[1,0] = x,y (y,x?) scales
    
    gang0,gr,gb=_angpos2(markers1)
    mang0,mr,mb=_angpos2(markers2)
    bartrans=gb-mb
    try: 
        logger
    except NameError:
        logger=logging.getLogger()
        logger.setLevel(logging.DEBUG)
    logger.info('- INPUT -' )
    logger.info("stdv of markers distance errors from barycenter: %s"%((gr-mr).std()))
    logger.info('translation: %s'%(bartrans))
    logger.info('- OUTPUT -' )  
    tmarkers=trans(markers1)
    tang0,tr,tb=_angpos2(tmarkers)
    logger.info("Affine Transformation Matrix:\n %s"%A)
    logger.info("stdv of markers distance errors from barycenter: %s"%((gr-tr).std()))
    logger.info('translation of barycenter: %s'%(gb-tb))
    err=markers2-trans(markers1)
    logger.info("errors in markers position after transformation: \n%s"%(err))
    logger.info("mean of error abs val: \n%s"%(np.sqrt((err**2).sum(axis=1)).mean()))
    
    return apply2D(trans)  


if __name__=="__main__":
    primary = np.array([[40., 1160., 0.],
                        [40., 40., 0.],
                        [260., 40., 0.],
                        [260., 1160., 0.]])

    secondary = np.array([[610., 560., 0.],
                          [610., -560., 0.],
                          [390., -560., 0.],
                          [390., 560., 0.]])
    
    print("-----------")
    print("Test affine")             
    print("-----------")         
    transform,A= find_affine(primary, secondary)
    print("Starting points:")
    print(primary)
    print("Target:")
    print(secondary)
    print("Result:")
    print(transform(primary))
    print("Max error:", np.abs(secondary - transform(primary)).max())
    
    
    print("-----------")
    print("Test rototrans")   
    print("-----------")
    secondary2=rotate_points(primary,math.pi/6,[-10,2])
    transform2,A2= find_rototrans(primary, secondary2,verbose=True)
    print("Result:")
    print(transform2(primary))
    

    rect1=np.array([[30,1200],[30,-600],[650,-600],[650,1200],[30,1200]])
    
    plt.figure()
    plot_transform([primary,secondary,rect1],[0,0,1],['markers1','markers2','ROI'],[transform2,None,transform2])
    plt.show()
    display(plt.gcf())
   



import numpy as np
import matplotlib.pyplot as plt
import scipy.optimize
import pdb
import utilities.imaging.man as man
import utilities.imaging.fitting as fit
import scipy.ndimage as nd
from utilities.plotting import scatter3d
import utilities.plotting
from utilities.imaging.analysis import rms
from scipy.interpolate import griddata

#This module contains routines to fit a cylinder to
#2D metrology data

def cyl(shape,curv,rad,yaw,pitch,roll,piston):
    """Create a cylindrical surface on a 2D array.
    Specify shape of array, and other parameters in
    pixels or radians where appropriate.
    Radius is assumed to be large enough to fill
    provided array.
    Curv is +1 or -1, with +1 indicating a convex
    cylinder from user's point of view (curving negative)"""
    #Construct coordinate grid
    x,y = np.meshgrid(np.arange(shape[1]),np.arange(shape[0]))
    x,y = x-np.mean(x),y-np.mean(y)
    #Apply yaw rotation
    x,y = x*np.cos(yaw)+y*np.sin(yaw),-x*np.sin(yaw)+y*np.cos(yaw)
    #Create base cylinder
    cyl = np.sqrt(rad**2-x**2)*curv
    #Apply pitch, roll, and piston
    cyl = cyl + piston + x*roll + y*pitch

    return cyl

def findGuess(d):
    """Find initial guess parameters for cylindrical
    metrology data. Use a quadratic fit in each axis
    to determine curvature sign and radius.
    Assume zero yaw"""
    sh = np.shape(d)
    yaw = 0.
    #Fit x and y slices
    xsl = d[sh[0]/2,:]
    ysl = d[:,sh[1]/2]
    xsl,ysl = xsl[np.invert(np.isnan(xsl))],ysl[np.invert(np.isnan(ysl))]
    Nx = np.size(xsl)
    Ny = np.size(ysl)
    xfit = np.polyfit(np.arange(Nx),xsl,1)
    yfit = np.polyfit(np.arange(Ny),ysl,1)
    #Subtract off linear fit
    xsl = xsl - np.polyval(xfit,np.arange(Nx))
    ysl = ysl - np.polyval(yfit,np.arange(Ny))
    #Figure out largest radius of curvature and sign
    Xsag = xsl.max()-xsl.min()
    Ysag = ysl.max()-ysl.min()
    if Xsag > Ysag:
        radius = Nx**2/8./Xsag
        curv = -np.sign(np.mean(np.diff(np.diff(xsl))))
    else:
        radius = Ny**2/8./Ysag
        curv = -np.sign(np.mean(np.diff(np.diff(ysl))))
        yaw = -np.pi/2
        
    return curv,radius,yaw,np.arctan(yfit[0]),\
                           np.arctan(xfit[0]),\
                           -radius+np.nanmean(d)

def fitCyl(d):
    """Fit a cylinder to the 2D data. NaNs are perfectly fine.
    Supply guess as [curv,rad,yaw,pitch,roll,piston]
    """
    guess = findGuess(d)
    fun = lambda p: np.nanmean((d-cyl(np.shape(d),*p))**2)
##    guess = guess[1:]
    pdb.set_trace()
    res = scipy.optimize.minimize(fun,guess,method='Powell',\
                    options={'disp':True,'ftol':1e-9,'xtol':1e-9})

    return res

def transformCyl(x,y,z,yaw,pitch,lateral,piston):
    """Transform x,y,z coordinates for cylindrical fitting"""
    #Yaw
    y,z = y*np.cos(yaw)+z*np.sin(yaw), -y*np.sin(yaw)+z*np.cos(yaw)
    #Pitch
    x,y = x*np.cos(pitch)+y*np.sin(pitch), -x*np.sin(pitch)+y*np.cos(pitch)
    #Lateral
    x = x + lateral
    #Piston
    z = z + piston
    return x,y,z

def itransformCyl(x,y,z,yaw,pitch,lateral,piston):
    """Transform x,y,z coordinates for cylindrical fitting"""
    #Lateral
    x = x + lateral
    #Piston
    z = z + piston
    #Pitch
    x,y = x*np.cos(-pitch)+y*np.sin(-pitch), -x*np.sin(-pitch)+y*np.cos(-pitch)
    #Yaw
    y,z = y*np.cos(-yaw)+z*np.sin(-yaw), -y*np.sin(-yaw)+z*np.cos(-yaw)

    return x,y,z

def meritFn(x,y,z):
    """Merit function for 3d cylindrical fitting"""
    rad = np.sqrt(x**2 + z**2)
    return rms(rad)

def fitCyl3D(d):
    """Fit a cylinder to the 2D data. NaNs are fine. Image is
    first unpacked into x,y,z point data. Use findGuess to
    determine initial guesses for cylindrical axis.
    """
    #Get cylindrical axis guess
    g = findGuess(d)
    #Convert to 3D points
    sh = np.shape(d)
    x,y,z = man.unpackimage(d,xlim=[-sh[1]/2.+.5,sh[1]/2.-.5],\
                            ylim=[-sh[0]/2.+.5,sh[0]/2.-.5])
    xf,yf,zf = man.unpackimage(d,xlim=[-sh[1]/2.+.5,sh[1]/2.-.5],\
                            ylim=[-sh[0]/2.+.5,sh[0]/2.-.5],remove=False)
    #Apply transformations to get points in about the right
    #area
    z = z - np.mean(z)
    x,y = x*np.cos(g[2])+y*np.sin(g[2]), -x*np.sin(g[2])+y*np.cos(g[2])
    y,z = y*np.cos(g[3])+z*np.sin(g[3]), -y*np.sin(g[3])+z*np.cos(g[3])
    x,z = x*np.cos(g[4])+z*np.sin(g[4]), -x*np.sin(g[4])+z*np.cos(g[4])
    z = z + g[0]*g[1]
    #Now apply fit to minimize sum of squares of radii
    fun = lambda p: meritFn(*transformCyl(x,y,z,*p))
    res = scipy.optimize.minimize(fun,[.01,.01,.01,.01],method='Powell',\
                    options={'disp':True})
    x,y,z = transformCyl(x,y,z,*res['x'])
    #Subtract mean cylinder
    rad = np.mean(np.sqrt(x**2+z**2))
    z = z - np.sqrt(rad**2-x**2)
    #Repack into 2D array
    yg,xg = np.meshgrid(np.arange(y.min(),y.max()+1),\
                        np.arange(x.min(),x.max()+1))
    newz = griddata(np.transpose([x,y]),z,\
                    np.transpose([xg,yg]),method='linear')
    pdb.set_trace()
##    xg,yg = np.meshgrid(np.linspace(-sh[1]/2.+.5,sh[1]/2.-.5,sh[1]),\
##                        np.linspace(-sh[0]/2.+.5,sh[0]/2.-.5,sh[0]))
##    pdb.set_trace()
##    newz = griddata(np.transpose([x,y]),z,\
##                    np.transpose([xg,yg]),method='linear')

    return newz



import os
from dataIO.fn_add_subfix import fn_add_subfix
from dataIO.span import span
from pySurf.points import get_points, resample_grid,rebin_points,subtract_points,plot_points,save_points
import numpy as np
from matplotlib import pyplot as plt

def FEAreader1(FEAfile):
    """read V. Marquez files. Perform column shifting and knvert x returning result in operator coordinates
    with all units in mm.
    """
    a=np.genfromtxt(FEAfile,delimiter=',',skip_header=1)
    yy,xx,zz,dy,dx,dz=np.hsplit(a[:,2:8],6)
    x=-xx-dx
    y=yy+dy  #Correct for gravity
    z=dz  #here I keep only the deformation, and ignore nominal cylinder shape
    p=np.hstack([x,y,dz])*1000. #convert to mm, all coordinates
    return p

def plotFEA(FEAfile,FEAreader,datafile=None,outname=None,markers=None):
    """ read FEA and resample/rebin both points and FEA on the grid defined by steps, subtract gravity from pts 
    and plot and returns the corrected points. Simulation and data are also plotted if valid. 
    If datafile is not passed only simulation data are plotted and returned.
    FEA reader is a function that accept a single argument FEAfile and return an array of points describing 
    changes in shape (x,y,dz)."""
    rebin=False
    
    fpts=FEAreader(FEAfile)    
    if trans is not None:
        fpts=trans(fpts)
    if  datafile is not None:
        pts=get_points(datafile,delimiter=' ')
        
        if rebin:
            pts=rebin_points(pts,steps=steps)
            fp=rebin_points(fpts,steps=steps)
        else:
            xl,yl=span(pts,axis=0)[:2,:]
            xg=np.arange(xl[0],xl[1],steps[0])
            yg=np.arange(yl[0],yl[1],steps[1])
            pts=resample_grid(pts,xg,yg)
            fp=resample_grid(fpts,xg,yg)

        pts0=subtract_points(pts,fp)
    else:
        pts0=pts
        pts=None
    
    #diagnostic plots for gravity subtraction
    if pts is not None:
        plt.figure('pts0')  #corrected
        plt.clf()
        plot_points(pts0,scatter=1,aspect='equal')
        plt.title('points from file')
        if markers is not None:
            plt.plot(markers[:,0],markers[:,1],'+w',label='support')

    if datafile is not None:        
        plt.figure('pts')  #data
        plt.clf()
        plot_points(pts,scatter=1,aspect='equal')
        plt.title('resampled and gravity corrected')
        plt.savefig(fn_add_subfix(outname,'_corrected','.png'))
        if markers is not None:
            plt.plot(markers[:,0],markers[:,1],'+w',label='support')
        
    plt.figure('fp')  #simulation
    plt.clf()
    plot_points(fp,scatter=1,aspect='equal')
    plt.title('gravity effect from FEA')
    if markers is not None:
        plt.plot(markers[:,0],markers[:,1],'+',label='support') #here x and y inverted
    plt.savefig(fn_add_subfix(outname,'_FEA','.png'))

    if outname is not None:
        save_points(pts0,fn_add_subfix(outname,'','.dat'))
        save_points(pts0,fn_add_subfix(outname,'_matrix','.dat'),matrix=1)
        save_points(pts0,fn_add_subfix(outname,'_matrix','.fits'),matrix=1)

    return pts0


from pySurf.points import *
from dataIO.fn_add_subfix import fn_add_subfix

def tomg_file(f):
    """convert matrix data file to 'smart' format martagiorgia.
    See format specs at the end.
    """
    p=get_points(f,delimiter=' ')
    dx,dy=points_find_grid(p,'step')[1]
    pp=matrix_to_points2(points_autoresample(p)[0])
    pp=pp[pp[:,0].argsort(kind='mergesort'),:]*[1,1,1000.]
    save_points(fn_add_subfix(f,'','.mgf'),pp,fill_value='32767',
        header='\n'.join([str(dx),str(dy)]),newline='\n',
        delimiter='\n',comments='')

def tomg(f,p):
    """convert points [Nx3] data to 'smart' format martagiorgia.
    See format specs at the end.
    """
    #p=get_points(f,delimiter=' ')
    dx,dy=points_find_grid(p,'step')[1]
    pp=matrix_to_points2(points_autoresample(p)[0])
    pp=pp[pp[:,0].argsort(kind='mergesort'),:]*[1,1,1000.]
    save_points(fn_add_subfix(f,'','.mgf'),pp,fill_value='32767',
        header='\n'.join([str(dx),str(dy)]),newline='\n',
        delimiter='\n',comments='')
    
if __name__=="__main__":
    file=r'170922_PCO1.2S03_RefSub_4in.dat'
    tomg_file(file)
    
'''
--------------------------------------------------------------------------------------------------
Questo formato e' adatto per passare sia superfici che funzioni di rimozione che matrici dei tempi
--------------------------------------------------------------------------------------------------

1) all'inizio del file mettere due righe contenenti il sampling in x e il sampling in y.

2) formato a 1 colonna, con rispettivamente in sequenza x,y,z. Poi si ripete per ogni punto successivo

3) i valori di x ed y sono interi, uguali agli indici dell'array ed iniziano entrambi da 0 

4) il valore di y varia piu' velocemente di x

5) alla fine di ogni riga mettere un carriage return + un line feed

6) il valore di z e' espresso in:
	- nanometri per le superfici (3 cifre decimali)
	- nanometri/sec per le funzioni di rimozione (tre cifre decimali)
	- secondi per le time matrix (3 cifre decimali)

7) laddove non c'e' informazione usare il valore 32767 come BAD


NOTA: se non si riesce a leggere il file e' possibile che non sia corretto il formato di fine riga che vuole un CRLF.
In tal caso un modo veloce di verificarlo e correggerlo e' tramite il programma Notepad++ che ha comandi specifici al caso citato


----------------------------------------------------------
esempio di file: (sampling di 2.249 in x e 2.249 in y)
----------------------------------------------------------


2.2490
2.2490
0.000	(spiegazione: x)
0.000	(y)
0.000	(z)
0.000	(x)
1.000	(y)
0.000	(z)
0.000	(x)
2.000	(y)
0.000	(z)
0.000	ecc........
3.000
0.000
0.000

'''


#2018/11/02 moved to scripts.repeatabilty from WFS_repeatability in POOL\PSD\WFS_repeatability
#v4 2018/09/18 include functions from reproducibility plot. Includes transformations
#  and differences for any number of plots (it was 3).
#v3 added functions all functions from repeatability.
#v1 works, this v2 adds finer control of styles in make_styles with list value:symbol.

import collections

import numpy as np
import matplotlib.pyplot as plt
import os
from cycler import cycler
from itertools import cycle
import pdb
import matplotlib.dates as mdates
import matplotlib.patches as mpatches
from matplotlib.lines import Line2D
import json
import pandas as pd
from plotting.backends import maximize
from pySurf.data2D_class import Data2D
from pySurf.readers.instrumentReader import fitsWFS_reader
from dataIO.fn_add_subfix import fn_add_subfix
from utilities.imaging import fitting as fit
from dateutil.parser import parse

from dataIO.outliers import remove_outliers
from plotting.add_clickable_markers import add_clickable_markers2
from pySurf.affine2D import find_affine, find_rototrans
from plotting.multiplots import find_grid_size, subplot_grid
import itertools

from IPython.display import display

#from pySurf.data2D_class import align_interactive

#from config.interface import conf_from_json

from pySurf.scripts.dlist import plot_data_repeat 
from pySurf.scripts.dlist import dcouples_plot   

"""2019/04/08 from messy script/repeatability, here sorted function that have been used. When a function is needed by e.g. a notebook, move it here from repeatability_dev.

Each script here should have parameters:
    dis :       call display after plots (and/or prints)
    outfile/outfolder :   produce output with this root name/folder 
    """

def removemis(D2D,func):
    """ convenience function.
    return a copy of data2D object D2D subtracting the results of a function 2D->2D applied to data.
    Original object is not modified."""
    res=D2D.copy() #avoid modification inplace
    res.data=res.data-func(res.data)[0]
    return res

    
def plot_data_repeat_leveling(dlist,outfile=None,dis=True,name = ""):
    """Create a multiplot on a grid with plots of all data with
    same leveling: raw, conical, cyliindrical, sag removed by line
    
    abused function in old notebooks when type of leveling was uncertain or of interest."""

    xs,ys=find_grid_size(len(dlist),3,square=False)
    res=[]
    r=plot_data_repeat(dlist,name=name,num=1)
    res.append(r)
    if outfile:
        os.makedirs(os.path.dirname(outfile)+'\\raw\\',exist_ok=True)
        plt.savefig(fn_add_subfix(outfile,'_raw','.png',pre='raw\\'))
    if dis: display(plt.gcf())
    
    fig,axes=plt.subplots(xs,ys,num=2)
    axes=axes.flatten()
    maximize()
    res.append([])
    for i,(ll,ax) in enumerate(zip(dlist,axes)):
        plt.subplot(xs,ys,i+1,sharex=axes[0],sharey=axes[0])
        tmp=ll.copy()
        tmp.data=ll.data-fit.fitCylMisalign(ll.data)[0]
        tmp.plot(stats=True)
        res[-1].append(tmp.std())
        #plt.clim([-3,3])
        plt.clim(*(tmp.std()*np.array([-1,1])))
    for ax in axes[:len(dlist)-1:-1]:
        fig.delaxes(ax)
        #plt.pause(0.1)
    plt.suptitle(name+' CYL corrected')
    if outfile:
        os.makedirs(os.path.dirname(outfile)+'\\cyl\\',exist_ok=True)
        plt.savefig(fn_add_subfix(outfile,'_cyl','.png',pre='cyl\\'))
    if dis: display(plt.gcf())
    
    fig,axes=plt.subplots(xs,ys,num=3)
    axes=axes.flatten()
    maximize()
    res.append([])
    for i,(ll,ax) in enumerate(zip(dlist,axes)):
        plt.subplot(xs,ys,i+1,sharex=axes[0],sharey=axes[0])
        tmp=ll.copy()
        tmp.data=ll.data-fit.fitConeMisalign(ll.data)[0]
        tmp.plot(stats=True)
        res[-1].append(tmp.std())
        #plt.clim([-3,3])
        plt.clim(*(tmp.std()*np.array([-1,1])))
    for ax in axes[:len(dlist)-1:-1]:
        fig.delaxes(ax)
        #plt.pause(0.1)
    plt.suptitle(name+' CONE corrected')
    if outfile:
        os.makedirs(os.path.dirname(outfile)+'\\cone\\',exist_ok=True)
        plt.savefig(fn_add_subfix(outfile,'_cone','.png',pre='cone\\'))
    if dis: display(plt.gcf())    
    
    fig,axes=plt.subplots(xs,ys,num=4)
    axes=axes.flatten()
    maximize()
    res.append([])
    for i,(ll,ax) in enumerate(zip(dlist,axes)):
        plt.subplot(xs,ys,i+1,sharex=axes[0],sharey=axes[0])
        tmp=ll.copy()
        #tmp.data=level_data(*ll(),2)[0]
        tmp=tmp.level(2,axis=1)
        tmp.plot(stats=True)
        res[-1].append(tmp.std())
        #plt.clim([-1,1])
        plt.clim(*(tmp.std()*np.array([-1,1])))
    for ax in axes[:len(dlist)-1:-1]:
        fig.delaxes(ax)
        #plt.pause(0.1)
    plt.suptitle(name+' SAG removed by line')
    plt.tight_layout(rect=[0,0,1,0.95])
    
    if outfile:
        os.makedirs(os.path.dirname(outfile)+'\\leg\\',exist_ok=True)
        plt.savefig(fn_add_subfix(outfile,'_leg','.png',pre='leg\\'))
    if dis: display(plt.gcf())  
        
    return res


def plot_repeat(rfiles,outfile=None,dis=True,name = "",plot_func=plot_data_repeat_leveling,ro=None):
    """Functions to plot a list of files side to side (e.g. repeatability or reproducibility)
    with different levelings. Return list of Data2D objects.
    plot_func is a function that accepts a dlist and possibly accepts arguments
       outfile, dis and name (it can have them ignored in **kwargs, or not passed at all to plot_repeat).
    This way, becomes a wrapper around plot_func (that doesn't necessarily have to plot,
      just have an interface outfile, dis, name.
    
    2019/04/08 made function general.
    2018/11/02 moved to scripts. Modified to make it format independent acting on
    data rather than on file extracting the data part in outer routine plot_data_repeat_leveling
    in ."""
    
    plt.close('all')

    if ro is None:
        ro={'reader':fitsWFS_reader,
                'scale':(-1,-1,1),
                'units':['mm','mm','um'],
                'ytox':220/200,
                'ypix':101.6/120,
                'center':(0,0)}
    
    #if name is None:
    #    name = os.path.basename(outfile) if outfile is not None else ""
    
    dlist=[Data2D(file=wf1,**ro).level() for i,wf1 in enumerate(rfiles)]
    res = plot_func(dlist,outfile=outfile,dis=dis,name = name)
    
    return dlist


    
from pySurf.readers.format_reader import auto_reader

    

    
    
def plot_rep_diff(dlist,outfile=None,dis=True):
    """Get three 2d arrays in a list. Calculate rotating differences for different component
    removal: plane, cylinder, cone, legendre.
    returns an 4 element list with the 4 possible removal for the 3 combinations of files to diff  """

    res=[]    
    
    plt.close('all')
    plt.figure(1)
    res.append(dcouples_plot(dlist))
    plt.suptitle('Differences RAW')
    if outfile is not None:
        plt.savefig(fn_add_subfix(outfile,'_raw','.png',pre='raw\\diff_'))
    if dis: display(plt.gcf())    
    
    plt.figure(2)
    res.append(dcouples_plot([removemis(dd,fit.fitCylMisalign) for dd in dlist]))
    plt.suptitle('Differences CYL removed')
    if outfile is not None:
        plt.savefig(fn_add_subfix(outfile,'_cyl','.png',pre='cyl\\diff_'))
    if dis: display(plt.gcf())    

    plt.figure(3)
    res.append(dcouples_plot([removemis(dd,fit.fitConeMisalign) for dd in dlist]))
    plt.suptitle('Differences CONE removed')
    if outfile is not None:
        plt.savefig(fn_add_subfix(outfile,'_cone','.png',pre='cone\\diff_'))
    if dis: display(plt.gcf())    

    plt.figure(4)
    res.append(dcouples_plot([dd.level((2,2)) for dd in dlist]))
    plt.suptitle('Differences 2,2 Legendre removed')
    if outfile is not None:
        plt.savefig(fn_add_subfix(outfile,'_leg','.png',pre='leg\\diff_'))
    if dis: display(plt.gcf())    
    
    return res



def psd_variability(psdlist,psdrange=None,rmsrange=None,fignum=None,units=None,
    outname=None,figsize=(15,5)):
    # to be removed
    """given a list of 2d psd files plot spans of all of them on same plot.
    TODO: replace input using data rather than files. Move to 2DPSD"""

    from plotting.fignumber import fignumber   
    from pySurf.data2D import projection
    from pySurf.psd2d import rms_power 
    from scipy.stats import ttest_ind
    
    units=['mm','mm','$\mu$m']
    labels = [p.name for p in psdlist]

    #gestione rudimentale dell'output troppo stanco per pensare allo standard.
    if outname is not None:
        label=os.path.basename(outname)
        outfolder=os.path.dirname(outname)
        os.makedirs(outfolder,exist_ok=True)

    rms_v=[]
    fig=fignumber(fignum,figsize=figsize)
    plt.subplot(121)
    for (p,x,f),fn in zip(psdlist,labels):
        ps=projection(p,axis=1,span=True)
        rms_v.append(rms_power(f,p,rmsrange))
        c=plt.plot(f,ps[0],label=os.path.basename(fn)+' AVG')[0].get_color()
        plt.plot(f,ps[1],'--',label='point-wise min',c=c)
        plt.plot(f,ps[2],'--',label='point-wise max',c=c)
        #pdb.set_trace()
    plt.ylabel('axial PSD ('+units[2]+'$^2$ '+units[1]+')')
    plt.xlabel('Freq. ('+units[1]+'$^{-1}$)')
    plt.vlines(rmsrange,*plt.ylim(),label='rms range',linestyle=':')
    plt.ylim(psdrange)
    plt.loglog()
    plt.grid(1)
    plt.legend(loc=0)

    print("Statistics for %s, rms range [%6.3g,%6.3g]:"%(label,*rmsrange))
    for rms,fn in zip(rms_v,labels):
        print(os.path.basename(fn)+": rms=%6.3g +/- %6.3g (evaluated on %i profiles)"%
            (np.nanmean(rms),np.nanstd(rms),len(rms)))
    print ("--------------")
    print ("Distribution of rms: %6.3g +/- %6.3g"%(np.nanmean(rms_v,axis=1).mean(),np.nanmean(rms_v,axis=1).std()))
    print ("AGGREGATED: rms=%6.3g +/- %6.3g (evaluated on %i profiles)"%
            (np.nanmean(rms_v),np.nanstd(rms_v),np.size(rms_v)))
    if len(rms_v)==2:
        print ("Student T test t:%6.3g p:%6.3g"%ttest_ind(rms_v[0],rms_v[1]))

    l=[os.path.splitext(os.path.basename(f))[0]+" rms=%3.2g +/- %3.2g (N=%i)"%
        (np.nanmean(r),np.nanstd(r),len(r)) for f,r in zip(labels,rms_v)]
    plt.subplot(122)
    plt.hist([rr[np.isfinite(rr)] for rr in rms_v],density=True,bins=100,label=l)
    plt.legend(loc=0)
    plt.title(label+" distrib. of avg: %3.2g +/- %3.2g"%
        (np.nanmean(rms_v,axis=1).mean(),np.nanmean(rms_v,axis=1).std()))
    #handles = [Rectangle((0,0),1,1,color=c,ec="k") for c in [low,medium, high]]
    #labels= ["low","medium", "high"]
    #plt.legend(handles, labels)
    plt.tight_layout()
    if outname is not None:
        plt.savefig(fn_add_subfix(outname,'_stats','.jpg'))

    return rms_v



#This submodule contains image analysis routines
from matplotlib.pyplot import *
from numpy import *
from matplotlib.colors import LogNorm
import pdb

def ptov(d):
    """Return the peak to valley of an image"""
    return nanmax(d)-nanmin(d)

def rms(d):
    """Return the RMS of an image"""
    return sqrt(nanmean((d-nanmean(d))**2))

def fitSag(d):
    """
    Compute sag to a vector by fitting a quadratic
    """
    if np.sum(np.isnan(d))==len(d):
        return np.nan
    x = np.arange(len(d))
    x = x[~np.isnan(d)]
    d = d[~np.isnan(d)]
    fit = np.polyfit(x,d,2)
    return fit[0]

def findMoments(d):
    x,y = meshgrid(arange(shape(d)[1]),arange(shape(d)[0]))
    cx = nansum(x*d)/nansum(d)
    cy = nansum(y*d)/nansum(d)
    rmsx = nansum((x-cx)**2*d)/nansum(d)
    rmsy = nansum((y-cy)**2*d)/nansum(d)
    pdb.set_trace()
    
    return cx,cy,sqrt(rmsx),sqrt(rmsy)

def nanflat(d):
    """
    Remove NaNs and flatten an image
    """
    d = d.flatten()
    d = d[invert(isnan(d))]
    return d

def fwhm(x,y):
    """Compute the FWHM of an x,y vector pair"""
    #Determine FWHM
    maxi = np.argmax(y) #Index of maximum value
    #Find positive bound
    xp = x[maxi:]
    fwhmp = xp[np.argmin(np.abs(y[maxi:]-y.max()/2))]-x[maxi]
    xm = x[:maxi]
    fwhmm = x[maxi]-xm[np.argmin(np.abs(y[:maxi]-y.max()/2))]
    return fwhmp+fwhmm

class pointGetter:
    """Creates an object tied to an imshow where the user can
    accumulate a list of x,y coords by right clicking on them.
    When done, the user can press the space key.
    """
    def __init__(self,img,log=False):
        ion()
        self.fig = figure()
        self.x = zeros(0)
        self.y = zeros(0)
        self.con = self.fig.canvas.mpl_connect('key_press_event',\
                                                self.keyEvent)
        ax = self.fig.add_subplot(111)
        if log is False:
            ax.imshow(img)
        else:
            ax.imshow(img,norm=LogNorm())
    def keyEvent(self,event):
        if event.key==' ':
            self.x = append(self.x,event.xdata)
            self.y = append(self.y,event.ydata)
    def close(self):
        self.fig.canvas.mpl_disconnect(self.con)
        close(self.fig)

class peakInformation:
    def __init__(self,img):
        self.img = img
        self.x0 = None
        self.y0 = None
        self.x1 = None
        self.y1 = None
        self.gx = []
        self.gy = []
        self.rmsx = []
        self.rmsy = []
        self.fig = figure()
        ax = self.fig.add_subplot(111)
        ax.imshow(img)
        self.con = self.fig.canvas.mpl_connect('button_press_event',self.clickEvent)
    def clickEvent(self,event):
        #If x0 and y0 are undefined, set them and return
        if self.x0 is None:
            #Define first point
            self.x0 = event.xdata
            self.y0 = event.ydata
            return
        #If x1 and y1 are undefined, set them and return centroid
        #Define second point
        self.x1 = event.xdata
        self.y1 = event.ydata
        #Order points properly
        x0 = min([self.x0,self.x1])
        x1 = max([self.x0,self.x1])
        y0 = min([self.y0,self.y1])
        y1 = max([self.y0,self.y1])            
        #Compute centroid between coordinates and update centroid list
        cx,cy,rmsx,rmsy = findMoments(self.img[y0:y1,x0:x1])
        print(('X: ' + str(cx+x0)))
        print(('Y: ' + str(cy+y0)))
        print(('RMS X: ' + str(rmsx)))
        print(('RMS Y: ' + str(rmsy)))
        try:
            self.gx.append(cx+x0)
            self.gy.append(cy+y0)
            self.rmsx.append(rmsx)
            self.rmsy.append(rmsy)
        except:
            self.gx = [cx+x0]
            self.gy = [cy+y0]
            self.rmsx = [rmsx]
            self.rmsy = [rmsy]
        #Reset event
        self.x0 = None
        self.x1 = None
        self.y0 = None
        self.y1 = None
    def close(self):
        self.fig.canvas.mpl_disconnect(self.con)
        print((self.gx))
        print((self.gy))
        print((self.rmsx))
        print((self.rmsy))
        close(self.fig)

def getPoints(img,log=False):
    """This function will bring up the image, wait for the user to
    press space while hovering over points, then wait for the user to
    end by pressing enter in command line, and then return a list of x,y coords
    """
    #Create instance of pointGetter class
    p = pointGetter(img,log=log)

    #When user presses enter, close the pointGetter class and
    #return the list of coordinates
    try:
        eval(input("Press enter when finished selecting points..."))
    except:
        pass
    x = p.x
    y = p.y
    p.close()

    return x,y

def getSubApp(img,log=False,points=None):
    """Return a subarray defined by rectangle enclosed by two points"""
    if points is None:
        x,y = getPoints(img,log=log)
    else:
        x,y = points
    return img[y.min():y.max(),x.min():x.max()]




"""Radial psf functions.

https://photutils.readthedocs.io/en/stable/psf.html

https://learn.astropy.org/rst-tutorials/synthetic-images.html?highlight=filtertutorials#3.-Prepare-a-Point-Spread-Function-(PSF)
"""


import numpy as np
from matplotlib import pyplot as plt
from dataIO.span import span


def barycenter(data,x,y):
    """return barycenter of 2D data, weighted by data."""
    xc = np.sum(x*data)/np.sum(data)
    yc = np.sum(y*data)/np.sum(data)

def radius (x,y,xc,yc):
    """calculate radius for points xx,yy from center in xc,yc.
    
    Return an array same size of xx/yy with all radii."""
    return np.sqrt((y-yc)**2+(x-xc)**2)

def radius_sorted(data,x,y,xc,yc):
    """return radius and corresponding data sorted by distance from xc,yc."""
    
    xg,yg = np.meshgrid(x,y)
    r = radius (xg.flatten(),yg.flatten(),xc,yc)
    i = np.argsort(r,axis=None)
    return r.flatten()[i],data.flatten()[i]

def radial_slice(x,y,angle,xyc,plot=False,
                    *args,**kwargs):
    """Gives extreme points of a segment passing by xc, yc with slope angle in radians inside the span of points x and y.
    works in all four quadrants."""
    
    if len(np.shape(angle))==0:
        angle = [angle]

    xx = x - xyc[0]
    yy = y - xyc[1]
    xs = span(xx)
    ys = span(yy)
    
    xxs = []
    for a in angle:
        cx = np.cos(a)
        cy = np.sin(a)

        t = xs/cx #parameter that brings to the x limits
        t2 = ys/cy #parameter that brings to the y limits
        # they are one negative and one positive if center in rectangle.
        tmin = max(min(t),min(t2))
        tmax = min(max(t),max(t2))
        p1 = (xyc[0]+tmin*cx,xyc[1]+tmin*cy)
        p2 = (xyc[0]+tmax*cx,xyc[1]+tmax*cy)
        xxs.append((p1,p2))
        
        if plot:
            l = plt.plot(*zip(p1,p2),*args,**kwargs)
            c = l[-1].get_color()
            plt.plot(*p1,'x',color = c)
            plt.plot(*p2,'^', color = c)
            #plt.arrow(*p1,*(np.array(p2)-np.array(p1)),*args,**kwargs)
    plt.show()
        
    return np.array(xxs)


"""functions to adjust linestyles."""

import matplotlib.pyplot as plt
import numpy as np


def colors_20(plot=False):
    """
    Build a 
    
    Lazily copied from:    https://stackoverflow.com/questions/8389636/creating-over-20-unique-legend-colors-using-matplotlib
    """
    
    NUM_COLORS = 20
    LINE_STYLES = ['solid', 'dashed', 'dashdot', 'dotted']
    NUM_STYLES = len(LINE_STYLES)

    colors = []
    styles = []
    cm = plt.get_cmap('gist_rainbow')


    for i in range(NUM_COLORS):
        lc = cm(i//NUM_STYLES*float(NUM_STYLES)/NUM_COLORS)
        ls = LINE_STYLES[i%NUM_STYLES]
        colors.append(lc)
        styles.append(ls)

    if plot:
        fig = plt.figure()
        ax = fig.add_subplot(111)
        data = [np.arange(10)*(i+1) for i in range(NUM_COLORS)]
        for i,(lc,ls) in enumerate(zip(colors,styles)):
            lines = ax.plot(data[i])
            lines[0].set_color(lc)
            lines[0].set_linestyle(ls)
        plt.show()
    
    return colors, styles


# -*- coding: utf-8 -*-

import numpy as np
import matplotlib.pyplot as plt
from dataIO.span import span

class Scale(object):
    '''
    2018/12/10 N.B.: this almost works, but scale is not accurate when zooming.
    probably something with how the length is rounded. use grid for comparison to 
    see the effect.
    
    A scale that can be plotted as line and text.
    Can be turned on and off setting the `visible` property. 
    .pixsize is the size of a pixel in the units used for the caption.
    The string from the unit is in self.un. 
    The method draw draws the scale if self.visible is True and 
    remove it if not. Instances of line and text are returned 
    (and can be modified, remember to call plt.draw to apply changes).
    Note that scale is independent on image, so it is plotted on current
    axis, it is care of the caller to set that.
    The scale updates itself on the plot on zooming, through callback
    connection through ax.callbacks.connect
    
    Vincenzo Cotroneo vcotroneo@cfa.harvard.edu 2016/03/07
    '''
    
    
    def __init__(self,pixsize=None,un="",visible=False):
        self._scaleFraction=0.25 #fraction of image used by scale 
        self._scaleRounding=1.1 #factor for rounding of scale, typically 1.1 for 10% works 
        self._scaleMargin= 0.05 #fraction of image to keep as distance from border in drawing the scale
        self.visible=visible
        self.pixsize=pixsize
        self.un=un
        self.l=None   #matplotlib line object
        self.t=None   #matplotlib text object
        ax=plt.gca()
        ax.callbacks.connect('xlim_changed', self.draw)
        if self.visible: self.draw()
   
    def draw(self,ax=None, pixsize=None):
        """from size of pixel pixsize (default to self.pixsize) 
        and corresponding real size imsize calculate a a suitable size for a 
        scale mark and the corresponding value and draw it. 
        The text for the unit is provided in un.
        Note that only matplotlib objects are used, an image is not necessary 
        and there are no pixels: pixsize represents the physical size of a unit of 
        axis coordinate (it matches pixel size if an image is plotted with pixel 
        coordinates, as default in imshow)  """
        
        if ax is None:
            ax=plt.gca()
        
        scaleFraction=self._scaleFraction #fraction of image used by scale 
        scaleRounding=self._scaleRounding #factor for rounding of scale, typically 1.1 for 10% works
        scaleMargin=  self._scaleMargin  #fraction of image to keep as distance from border in drawing the scale
        
        #pixsize is set to the correct value:
        if not (pixsize is None): 
            self.pixsize=pixsize
        if self.pixsize is None:
            pixsize=1.
        else: pixsize=self.pixsize
        imsize=(span(ax.get_xlim(),size=1),span(ax.get_ylim(),size=1))
        scalesizeunits=int(round(scaleRounding*imsize[0]*scaleFraction*pixsize)) #size of the scale in units of un
        scalesizepx=scalesizeunits/pixsize   #size of the scale in pixels
        #determine position on plot
        ##top right 

        #xpos=plt.xlim()[0]+imsize[0]*(1-scaleMargin)-scalesizepx
        #ypos=imsize[1]*scaleMargin
        scaleFraction=scalesizepx/imsize[0] #update to correct the rounding
        xpos=1-scaleFraction-scaleMargin
        ypos=1-scaleMargin
        scaletxt="%5.3f %s"%(scalesizeunits,self.un)
        
        if not(self.l is None):
            lcol=self.l.get_color()  #ugly solution to keep color (but not other props)
            self.l.remove()
            self.l=None
        else:
            lcol='yellow'
        if not(self.t is None):
            tcol=self.t.get_color()  #ugly solution to keep color (but not other props)
            self.t.remove()
            self.t=None
        else:
            tcol='yellow'
        if self.visible:
            aux=ax.get_autoscalex_on()
            auy=ax.get_autoscaley_on()
            ax.set_autoscale_on(False)
            #self.l=plt.plot([xpos,xpos+scalesizepx],[ypos,ypos],'y',lw=3)[0]
            #self.t=plt.text(xpos,ypos+imsize[0]*scaleMargin,scaletxt,color='yellow')  
            self.l=ax.plot([xpos,xpos+scaleFraction],[ypos,ypos],'y',lw=3,transform=ax.transAxes,color=tcol)[0]
            self.t=ax.text(xpos,(ypos-scaleMargin),scaletxt,transform=ax.transAxes,color=lcol)
            aux=ax.set_autoscalex_on(aux)
            auy=ax.set_autoscaley_on(auy)
            
        plt.draw()

    def remove(self):
        v=self.visible
        self.visible=False
        self.draw()
        self.visible=v
    
        return self.l,self.t


def test_with_image(infile,outfolder=None):
    """apply a scale on a image. Image can be aligned and zoomed to verify scale
    and scale update."""
    
    infolder=os.path.dirname(infile)
    imgfile=os.path.basename(infile)    
    img=plt.imread(infile)
    
    #plot green channel
    plt.clf()
    plt.imshow(img[:,:,1])
    
    #fix pixsize: 300 um is 810 units of axis (pixels).
    #visible is set, scale should be immediately plotted.
    s=Scale(300/810.,'um',True) #810 pixel for 30 small ruler units, each unit is 10 um. pixsize is 300/810.
    
    if outfolder is not None:
        plt.savefig(os.path.join(outfolder,fn_add_subfix(imgfile,'_01','.png')))
    
    #zoom 2x 
    plt.xlim(np.array(plt.xlim())//2)
    plt.ylim(np.array(plt.ylim())//2)
    s.draw()
    if outfolder is not None:
            plt.savefig(os.path.join(outfolder,fn_add_subfix(imgfile,'_02','.png')))
    return s    

def test_with_plot(outfolder=None):
    """Apply scale to a line plot, it should work the same as an image."""
    
    N=100  #nr. of points in plot
    L=20.  #length of x axis
    x=np.arange(N)*L/(N-1)
    y=np.random.random(N)
    plt.clf()
    plt.plot(x,y)
    
    #fix pixsize: 300 um is 810 units of axis (pixels).
    #visible is set, scale should be immediately plotted.
    s=Scale(L/span(plt.xlim(),True),'units of x',True) #The entire axis span is L
    
    if outfolder is not None:
        plt.savefig(os.path.join(outfolder,'test_with_plot_01.png'))
    
    #zoom 3x 
    #plt.xlim(plt.xlim()[0]/3.,plt.xlim()[1]/3)
    #plt.ylim(plt.ylim()[0]/3.,plt.ylim()[1]/3)    
    s.t.set_color('black')
    s.l.set_color('black')
    s.draw()
    
    plt.grid(1)
    
    if outfolder is not None:
            plt.savefig(os.path.join(outfolder,'test_with_plot_02.png'))
    return s    

    
if __name__=='__main__':
    from dataIO.fn_add_subfix import fn_add_subfix
    import os
    #plt.ion()
    
    infolder=r'test\input_data'
    outfolder=r'test\Scale'
    imgfile='03_calib_20x.jpg'  #img of a ruler at 20x

    plt.close('all')
    
    # this partially fails: 1) color is set and drawn, but reset to yellow on 
    # scale.draw(); 
    # 2) starts from the zoomed version, that is considered home in zoom history even if the image/plot is larger. 
    # 3) makes wrong placement of scale on strong zoom.
    s=test_with_image(os.path.join(infolder,imgfile),outfolder)
    plt.show(block=True) #this goes in main or in scripts that are called by command line

    #this also 
    plt.figure()
    s2=test_with_plot(outfolder)
    s2.t.set_color('black')
    s2.l.set_color('black')
    plt.show(block=True) #this goes in main or in scripts that are called by command line    
   


import numpy as np
import pdb
from scipy.interpolate import griddata
from imaging.man import stripnans,nearestNaN
from scipy.integrate import simps

#This module contains Fourier analysis routine

def components(d,win=np.hanning):
    """Want to return Fourier components with optional window
    Application note: These components are dependent on sampling!
    This means you can *not* interpolate these components onto other
    frequency grids!
    """
    #Handle window
    if win is not 1:
        if np.size(np.shape(d)) is 1:
            win = win(np.size(d))/np.sqrt(np.mean(win(np.size(d))**2))
        else:
            win1 = win(np.shape(d)[0])
            win2 = win(np.shape(d)[1])
            win = np.outer(win1,win2)
            win = win/np.sqrt(np.mean(win**2))

    #Compute Fourier components
    return np.fft.fftn(d*win)/np.size(d)

def continuousComponents(d,dx,win=np.hanning):
    """Want to return Fourier components with optional window
    Divide by frequency interval to convert to continuous FFT
    These components can be safely interpolated onto other frequency
    grids. Multiply by new frequency interval to get to numpy format
    FFT. Frequency units *must* be the same in each case.
    """
    #Handle window
    if win is not 1:
        if np.size(np.shape(d)) is 1:
            win = win(np.size(d))/np.sqrt(np.mean(win(np.size(d))**2))
        else:
            win1 = win(np.shape(d)[0])
            win2 = win(np.shape(d)[1])
            win = np.outer(win1,win2)
            win = win/np.sqrt(np.mean(win**2))

    #Compute Fourier components
    return np.fft.fftn(d*win)*dx

def newFreq(f,p,nf):
    """
    Interpolate a power spectrum onto a new frequency grid.
    """
    return griddata(f,p,nf,method=method)

def freqgrid(d,dx=1.):
    """Return a frequency grid to match FFT components
    """
    freqx = np.fft.fftfreq(np.shape(d)[1],d=dx)
    freqy = np.fft.fftfreq(np.shape(d)[0],d=dx)
    freqx,freqy = np.meshgrid(freqx,freqy)
    return freqx,freqy

def ellipsoidalHighFrequencyCutoff(d,fxmax,fymax,dx=1.,win=np.hanning):
    """A simple low-pass filter with a high frequency cutoff.
    The cutoff boundary is an ellipsoid in frequency space.
    All frequency components with (fx/fxmax)**2+(fy/fymax)**2 > 1.
    are eliminated.
    fxmax refers to the second index, fymax refers to the first index
    This is consistent with indices in imshow
    """
    #FFT components in numpy format
    fftcomp = components(d,win=win)*np.size(d)

    #Get frequencies
    freqx,freqy = freqgrid(d,dx=dx)

    #Get indices of frequencies violating cutoff
    ind = (freqx/fxmax)**2+(freqy/fymax)**2 > 1.
    fftcomp[ind] = 0.

    #Invert the FFT and return the filtered image
    return fft.ifftn(fftcomp)

def meanPSD(d0,win=np.hanning,dx=1.,axis=0,irregular=False,returnInd=False,minpx=10):
    """Return the 1D PSD averaged over a surface.
    Axis indicates the axis over which to FFT
    If irregular is True, each slice will be stripped
    and then the power spectra
    interpolated to common frequency grid
    Presume image has already been interpolated internally
    If returnInd is true, return array of power spectra
    Ignores slices with less than minpx non-nans
    """
    #Handle which axis is transformed
    if axis==0:
        d0 = np.transpose(d0)
    #Create list of slices
    if irregular is True:
        d0 = [stripnans(di) for di in d0]
    else:
        d0 = [di for di in d0]
    #Create power spectra from each slice
    pows = [realPSD(s,win=win,dx=dx,minpx=minpx) for s in d0 \
            if np.sum(~np.isnan(s)) >= minpx]
    #Interpolate onto common frequency grid of shortest slice
    if irregular is True:
        #Determine smallest frequency grid
        ln = [len(s[0]) for s in pows]
        freq = pows[np.argmin(ln)][0]
        #Interpolate
        pp = [griddata(p[0],p[1],freq) for p in pows]
    else:
        pp = [p[1] for p in pows]
        freq = pows[0][0]
    #Average
    pa = np.mean(pp,axis=0)
    if returnInd is True:
        return freq,pp
    return freq,pa

def medianPSD(d0,win=np.hanning,dx=1.,axis=0,nans=False):
    """Return the 1D PSD "medianed" over a surface.
    Axis indicates the axis over which to FFT
    If nans is True, each slice will be stripped,
    internally interpolated, and then the power spectra
    interpolated to common frequency grid"""
    d = stripnans(d0)
    if win is not 1:
        win = win(np.shape(d)[axis])/\
              np.sqrt(np.mean(win(np.shape(d)[axis])**2))
        win = np.repeat(win,np.shape(d)[axis-1])
        win = np.reshape(win,(np.shape(d)[axis],np.shape(d)[axis-1]))
        if axis is 1:
            win = np.transpose(win)
    c = np.abs(np.fft.fft(d*win,axis=axis)/np.shape(d)[axis])**2
    c = np.median(c,axis=axis-1)
    f = np.fft.fftfreq(np.size(c),d=dx)
    f = f[:np.size(c)/2]
    c = c[:np.size(c)/2]
    c[1:] = 2*c[1:]
    return f,c

def realPSD(d0,win=np.hanning,dx=1.,axis=None,nans=False,minpx=10):
    """This function returns the PSD of a real function
    Gets rid of zero frequency and puts all power in positive frequencies
    Returns only positive frequencies
    """
    if nans is True:
        d = stripnans(d0)
    else:
        d = d0
    if len(d) < minpx:
        return np.nan
    #Get Fourier components
    c = components(d,win=win)
    #Handle collapsing to 1D PSD if axis keyword is set
    if axis==0:
        c = c[:,0]
    elif axis==1:
        c = c[0,:]

    #Reform into PSD
    if np.size(np.shape(c)) is 2:
        f = [np.fft.fftfreq(np.shape(c)[0],d=dx)[:np.shape(c)[0]/2],\
                   np.fft.fftfreq(np.shape(c)[1],d=dx)[:np.shape(c)[1]/2]]
        c = c[:np.shape(c)[0]/2,:np.shape(c)[1]/2]
        c[0,0] = 0.
        #Handle normalization
        c = 2*c
        c[0,:] = c[0,:]/np.sqrt(2.)
        c[:,0] = c[:,0]/np.sqrt(2.)
        
    elif np.size(np.shape(c)) is 1:
        f = np.fft.fftfreq(np.size(c),d=dx)
        f = f[:np.size(c)/2]
        c = c[:np.size(c)/2]
        c[0] = 0.
        c = c*np.sqrt(2.)

    return f[1:],np.abs(c[1:])**2

def computeFreqBand(f,p,f1,f2,df,method='linear'):
    """
    Compute the power in the PSD between f1 and f2.
    f and p should be as returned by realPSD or meanPSD
    Interpolate between f1 and f2 with size df
    Then use numerical integration
    """
    newf = np.linspace(f1,f2,(f2-f1)/df+1)
    try:
        newp = griddata(f,p/f[0],newf,method=method)
    except:
        pdb.set_trace()
    return np.sqrt(simps(newp,x=newf))

def fftComputeFreqBand(d,f1,f2,df,dx=1.,win=np.hanning,nans=False,minpx=10,\
                       method='linear'):
    """
    Wrapper to take the FFT and immediately return the
    power between f1 and f2 of a slice
    If slice length is < 10, return nan
    """
    if np.sum(~np.isnan(d)) < minpx:
        return np.nan
    f,p = realPSD(d,dx=dx,win=win,nans=nans)
    return computeFreqBand(f,p,f1,f2,df,method=method)

def psdScan(d,f1,f2,df,N,axis=0,dx=1.,win=np.hanning,nans=False,minpx=10):
    """
    Take a running slice of length N and compute band limited
    power over the entire image. Resulting power array will be
    of shape (S1-N,S2) if axis is 0
    axis is which axis to FFT over
    """
    if axis is 0:
        d = np.transpose(d)
    sh = np.shape(d)
    m = np.array([[fftComputeFreqBand(di[i:i+N],f1,f2,df,dx=dx,win=win,nans=nans,minpx=minpx) \
      for i in range(sh[1]-N)] for di in d])
    if axis is 0:
        m = np.transpose(m)
    return m
    

def lowpass(d,dx,fcut):
    """Apply a low pass filter to a 1 or 2 dimensional array.
    Supply the bin size and the cutoff frequency in the same units.
    """
    #Get shape of array
    sh = np.shape(d)
    #Take FFT and form frequency arrays
    f = np.fft.fftn(d)
    if np.size(np.shape(d)) > 1:
        fx = np.fft.fftfreq(sh[0],d=dx)
        fy = np.fft.fftfreq(sh[1],d=dx)
        fa = np.meshgrid(fy,fx)
        fr = np.sqrt(fa[0]**2+fa[1]**2)
    else:
        fr = np.fft.fftfreq(sh[0],d=dx)
    #Apply cutoff
    f[fr>fcut] = 0.
    #Inverse FFT
    filtered = np.fft.ifftn(f)
    return filtered

def randomizePh(d):
    """Create a randomized phase array that maintains a real
    inverse Fourier transform. This requires that F(-w1,-w2)=F*(w1,w2)
    """
    #Initialize random phase array
    sh = np.shape(d)
    ph = np.zeros(sh,dtype='complex')+1.
    
    #Handle 1D case first
    if np.size(sh) == 1:
        if np.size(d) % 2 == 0:        
            ph[1:sh[0]/2] = np.exp(1j*np.random.rand(sh[0]/2-1)*2*np.pi)
            ph[sh[0]/2+1:] = np.conjugate(np.flipud(ph[1:sh[0]/2]))
        else:
            ph[1:sh[0]/2+1] = np.exp(1j*np.random.rand(sh[0]/2)*2*np.pi)
            ph[sh[0]/2+1:] = np.conjugate(np.flipud(ph[1:sh[0]/2+1]))
    else:
        #Handle zero frequency column/rows
        ph[:,0] = randomizePh(ph[:,0])
        ph[0,:] = randomizePh(ph[0,:])
        #Create quadrant
        if sh[0] % 2 == 0 and sh[1] % 2 == 0:
            #Handle intermediate Nyquist
            ph[sh[0]/2,:] = randomizePh(ph[sh[0]/2,:])
            ph[:,sh[1]/2] = randomizePh(ph[:,sh[1]/2])
            #Form quadrant
            ph[1:sh[0]/2,1:sh[1]/2] = \
                np.exp(1j*np.random.rand(sh[0]/2-1,sh[1]/2-1)*2*np.pi)
            ph[sh[0]/2+1:,sh[1]/2+1:] = \
                np.conjugate(np.flipud(np.fliplr(ph[1:sh[0]/2,1:sh[1]/2])))
            ph[1:sh[0]/2,sh[1]/2+1:] = \
                np.exp(1j*np.random.rand(sh[0]/2-1,sh[1]/2-1)*2*np.pi)
            ph[sh[0]/2+1:,1:sh[1]/2] = \
                np.conjugate(np.flipud(np.fliplr(ph[1:sh[0]/2,sh[1]/2+1:])))
        elif sh[0] % 2 == 0 and sh[1] % 2 == 1:
            #Handle intermediate Nyquist
            ph[sh[0]/2,:] = randomizePh(ph[sh[0]/2,:])
            #Form quadrant
            ph[1:sh[0]/2,1:sh[1]/2+1] = \
                np.exp(1j*np.random.rand(sh[0]/2-1,sh[1]/2)*2*np.pi)
            ph[sh[0]/2+1:,sh[1]/2+1:] = \
                np.conjugate(np.flipud(np.fliplr(ph[1:sh[0]/2,1:sh[1]/2+1])))
            ph[1:sh[0]/2,sh[1]/2+1:] = \
                np.exp(1j*np.random.rand(sh[0]/2-1,sh[1]/2)*2*np.pi)
            ph[sh[0]/2+1:,1:sh[1]/2+1] = \
                np.conjugate(np.flipud(np.fliplr(ph[1:sh[0]/2,sh[1]/2+1:])))
        elif sh[0] % 2 == 1 and sh[1] % 2 == 0:
            #Handle intermediate Nyquist
            ph[:,sh[1]/2] = randomizePh(ph[:,sh[1]/2])
            #Form quadrant
            ph[1:sh[0]/2+1,1:sh[1]/2] = \
                np.exp(1j*np.random.rand(sh[0]/2,sh[1]/2-1)*2*np.pi)
            ph[sh[0]/2+1:,sh[1]/2+1:] = \
                np.conjugate(np.flipud(np.fliplr(ph[1:sh[0]/2+1,1:sh[1]/2])))
            ph[1:sh[0]/2+1,sh[1]/2+1:] = \
                np.exp(1j*np.random.rand(sh[0]/2,sh[1]/2-1)*2*np.pi)
            ph[sh[0]/2+1:,1:sh[1]/2+1] = \
                np.conjugate(np.flipud(np.fliplr(ph[1:sh[0]/2+1,sh[1]/2:])))
        else:
            #Form quadrant
            ph[1:sh[0]/2+1,1:sh[1]/2+1] = \
                np.exp(1j*np.random.rand(sh[0]/2,sh[1]/2)*2*np.pi)
            ph[sh[0]/2+1:,sh[1]/2+1:] = \
                np.conjugate(np.flipud(np.fliplr(ph[1:sh[0]/2+1,1:sh[1]/2+1])))
            ph[1:sh[0]/2+1,sh[1]/2+1:] = \
                np.exp(1j*np.random.rand(sh[0]/2,sh[1]/2)*2*np.pi)
            ph[sh[0]/2+1:,1:sh[1]/2+1] = \
                np.conjugate(np.flipud(np.fliplr(ph[1:sh[0]/2+1,sh[1]/2+1:])))
            
        
##        if np.size(d) % 2 == 1:
##            ph[1:sh[0]/2] = np.random.rand(sh[0]/2-1)*2*np.pi
##            pdb.set_trace()
##            ph[sh[0]/2+1:] = np.conjugate(np.flipud(ph[1:sh[0]/2]))
##        else:
##            ph[1:(sh[0]-1)/2] = np.random.rand((sh[0]-1)/2-1)*2*np.pi
##            pdb.set_trace()
##            ph[(sh[0]+1)/2:] = np.conjugate(np.flipud(ph[1:(sh[0]-1)/2]))

            
##    #Fill in positive x frequencies with random phases
##    ind = freqx >= 0.
##    ph[ind] = np.exp(1j*np.random.rand(np.sum(ind))*2*np.pi)
##    #Fill in negative x frequencies with complex conjugates
##    ph[np.ceil(sh[0]/2.):,0] = np.conjugate(\
##        np.flipud(ph[:np.floor(sh[0]/2.),0]))
##    ph[0,np.ceil(sh[1]/2.):] = np.conjugate(\
##        np.flipud(ph[0,:np.floor(sh[1]/2.)]))

    return ph
    
def randomProfile(freq,psd):
    """
    Generate a random profile from an input PSD.
    freq should be in standard fft.fftfreq format
    psd should be symmetric as with a real signal
    sqrt(sum(psd)) will equal RMS of profile
    """
    amp = np.sqrt(psd)*len(freq)
    ph = randomizePh(amp)
    f = amp*ph
    sig = np.fft.ifft(f)
    return np.real(sig)



from math import *

class fig(object):
    def __init__ (self,origin):
        self.origin=tuple(map(float,origin))
        object.__init__(self)
    def __repr__(self):
        #s="\n--<"+str(self.__class__) + " instance at " + hex(id(self)) + ">--\n"
        s=object.__repr__(self)+"\n"
        for k in list(vars(self).items()): s=s+ "%s = %s\n" %k
        s=s+"--------------------------\n\n"
        return s
    def __eq__(self,other):
        print("metodo di confronto non implementato")
        return None

class punto(fig):
    def __init__(self,x,y,z,origin=(0,0,0)):
        self.x=float(x)
        self.y=float(y)
        self.z=float(z)
        fig.__init__(self,origin)
    def trasla (self,new_origin):
        #print "translazione non ancora implementata"
        n_o=(self.origin[0]-new_origin[0],self.origin[1]-new_origin[1],self.origin[2]-new_origin[2])
        p=punto(self.x+n_o[0],self.y+n_o[1],self.z+n_o[2],new_origin)
        return p
    
class raggio(fig):
    ''' un raggio di luce e' una retta nel piano // a yz (x=origin[0])
    equazione fissata da due parametri m e q. m,q= None per retta verticale
    (// asse z, equazione x=origin[0], y=origin[1])'''
    def __init__(self,m,q,origin=(0,0,0)):
        if m!= None:
            self.m=float(m)
        else:
            self.m=None
        self.q=float(q)
        fig.__init__(self,origin)
    def trasla (self,new_origin):
        n_o=(new_origin[0]-self.origin[0],new_origin[1]-self.origin[1],new_origin[2]-self.origin[2])
        if self.m != None:
            new_q=self.q+self.m*n_o[1]-n_o[2]
            r=raggio(self.m,new_q,new_origin)
        else:
            r=raggio(None,self.q-n_o[1],new_origin)
        return r
    def getTh(self):
        return math.atan(1/self.m)
    def setTh(self, value):
        if value==math.pi/2:
            self.m=0
        elif value==0:
            self.m=None
        else:
            self.m = 1/math.tan(value)
    theta = property(getTh, setTh, "theta angle with z axis")

class proPlane(fig):
    ''' piano // a asse x con inclinazione rispetto asse z,
    uguale all'angolo off-axis. e' il piano su cui si proietta
    l'apertura dell'ottica. equazione fissata da due parametri m e q.
    '''
    def __init__(self,m,q,origin=(0,0,0)):
        self.m=float(m)
        self.q=float(q)
        fig.__init__(self,origin)
    def projectCircle(self,R,center):
        #da' R valore del raggio e center punto di centro
        center=center.trasla(self.origin)
        #retta perpendicolare al piano passante per il centro
        yc=center.y
        zc=center.z
        m=self.m
        q=self.q
        #r=raggio(-1/m,zc+yc/m,self.origin)
        #coordiate dell'intersezione col piano
        ncy=(zc+yc/m-q)/(m+(1/m))
        ncz=ncy*m+q
        new_cen=punto(center.x,ncy,ncz,self.origin)
        pass
        return new_cen
        
    def getTh(self):
        return math.atan(1/self.m)
    def setTh(self, value):
        if value==math.pi/2:
            self.m=0
        else:
            self.m = 1/math.tan(value)
    theta = property(getTh, setTh, "theta angle with z axis")
    
class hyp(fig) :
    '''iperbole determinata (nel piano) dai due parametri a e b,
    secondo l'equazione (z/a)^2-(y/b)^2=1 '''
    def __init__(self,a,b,origin=(0,0,0)):
        self.a=float(a)
        self.b=float(b)
        fig.__init__(self,origin)
    def rayIntersect(self,raggio):
        '''restituisce uno o piu' punti intersezione tra iperbole
        e raggio'''
        #parametri dell'equazione quadratica associata al sistema
        if self.origin[0] != raggio.origin[0]:print("attenzione! self.origin[0] != raggio.origin[0]:\n",\
            self.origin[0],raggio.origin[0])
        raggio=raggio.trasla(self.origin)
        if raggio.m != None:    
            a = (raggio.m/self.a)**2-(1/self.b)**2
            b = 2*raggio.m*raggio.q/self.a**2
            c = -1 + (raggio.q/self.a)**2
            #calcola delta
            delta = b**2-4*a*c
            if delta >= 0:
                #risolve formula
                solY=((-b-math.sqrt(delta))/(2*a),(-b+math.sqrt(delta))/(2*a))
                solZ=[math.sqrt((1+(y/self.b)**2)*self.a**2) for y in solY]
                p1=punto(0,solY[0],solZ[0],self.origin)
                p2=punto(0,solY[1],solZ[1],self.origin)
                return p1,p2
            else:
                return None
        else:
            solZ=math.sqrt((1+(raggio.q/self.b)**2)*self.a**2)
            return punto(0,raggio.q,solZ,self.origin)
        
    def tanInPmq(self,p):
        '''restituisce m e q del raggio tangente all'iperbole nel punto p
        nelle coordinate con origine self.origin'''
        p=p.trasla(self.origin)
        b=self.b
        a=self.a
        y0=p.y
        z0=p.z
        #controlla appartenenza
        if not(((y0/a)**2-(z0/b)**2)==1):
            print("punto non appartenente all'iperbole")
            #return None
        #calcola con formula
        q=(b**2)/z0
        m=(q*y0)/a**2
        return m,q
        
    def tanInP (self,p):
        '''restituisce il raggio tangente all'iperbole nel punto p'''
        '''simile a tanInP, restituisce l'angolo tra tangente e
        raggio incidente nel punto p '''
        m,q=tanInPmq(self,p)
        return raggio(m,q,self.origin)

    def tanInPang (self,p):
        '''restituisce l'angolo tra tangente e
        raggio incidente nel punto p '''
        m,q=tanInPmq(self,p)
        return math.atan(1/m)

class shell1(fig):
    ''' shell determinata da due parametri tra raggio, focale, angolo
    non considero l'altezza in quanto la prendo come unita' di misura,
    quindi H=1. considero solo la prima semisuperficie '''
    def __init__(self,R,F,origin=(0,0,0),limit=-1):
        self.F=float(F)
        self.alfa=atan2(R/F)/4
        fig.__init__(self,origin)
        if limit==0:        #non considero limiti: cono infinito
            self.limite=0 
        elif limit == -1: #limiti predefiniti 0 e 1
            self.limite=limite()
        else:
            self.limite=limit
    def planeIntersect (self,x0):
        '''intersezione con un piano // a yz, con eq. x=x0
        (rispetto all'origine standard)'''
        x0=x0-self.origin[0]
        h=hyp((x0/tan(self.alfa)),(xo),(0,0,-self.R/tan(alfa)),self.origin)
        return h
    def rayIntersect (self,raggio):
        ''' da' le coordinate del punto di intersezione con un raggio'''
        h=planeIntersect(raggio.origin[0])
        return h.rayIntersect(raggio)

class limite (fig):
    def __init__ (self,zmin=0.,zmax=1.,origin=(0,0,0)):
        '''limite della shell, un po' da definire'''
        self.zmin=float(zmin)
        self.zmax=float(zmax)
        fig.__init__(self,origin)



if __name__ == "__main__":
    pass

    o=punto(3,-2,1)
    pl=proPlane(-2,4)
    print(pl.projectCircle(5,o))

    '''
    #test rette verticali
    h=hyp(2,3)
    r=raggio(0,0)
    r.theta=0
    print h.rayIntersect(r)
    r.q=1
    print h.rayIntersect(r)'''
    
    ''' #test intersezione hyp-retta
    h=hyp(2.,math.sqrt(12.))
    r=raggio(1./3,2.)
    l=limite(1,3)
    p=list( h.rayIntersect(r))
    print p
    psel=[i for i in p if (l.zmin <= i.z <= l.zmax)]
    print psel    
    '''
    
    '''#test traslazione di un punto
    r=raggio(3,2)
    print "prima di traslazione r=raggio(3,2):"
    print r
    print "dopo traslazione r.trasla((0,0,2)):"
    print r.trasla((0,0,2))
    '''

    ''' #test tangente in un punto    
    h=hyp(2,3)
    p=punto(0,3.,math.sqrt(5)*1.5)
    print h.tanInP(p)
    '''


    


#!/usr/bin/env python
import numpy
import Image
import os

from astropy.io import fits

def jpgToFITS(filename):
	'''Extract the RGB channels and save them as FITS.'''
	#get the image and color information
	image = Image.open(filename)
	image.load()
	try:
		timestamp=image._getexif()[306]
	except:
		timestamp=''
	#306=datetime
	#36867: 'DateTimeOriginal',
	#36868: 'DateTimeDigitized',
	#33434: 'ExposureTime',
	#33437: 'FNumber',
	#PIL.ExifTags.Tags gives all the tag names
	#image.show()
	xsize, ysize = image.size
	r, g, b = image.split()
	rdata = r.getdata() # data is now an array of length ysize\*xsize
	gdata = g.getdata()
	bdata = b.getdata()

	# create numpy arrays
	npr = numpy.reshape(rdata, (ysize, xsize))
	npg = numpy.reshape(gdata, (ysize, xsize))
	npb = numpy.reshape(bdata, (ysize, xsize))

	# write out the fits images, the data numbers are still JUST the RGB
	# scalings; don't use for science
	red = fits.PrimaryHDU(data=npr)
	red.header['TIMESTAMP'] = timestamp # add spurious header info
	red.header['CHANNEL'] = "red"
	red.writeto(os.path.splitext(filename)[0]+'_r.fits',clobber=True)

	green = fits.PrimaryHDU(data=npg)
	green.header['TIMESTAMP'] = timestamp # add spurious header info
	green.header['CHANNEL'] = "green"
	green.writeto(os.path.splitext(filename)[0]+'_g.fits',clobber=True)

	blue = fits.PrimaryHDU(data=npb)
	blue.header['TIMESTAMP'] = timestamp # add spurious header info
	blue.header['CHANNEL'] = "blue"
	blue.writeto(os.path.splitext(filename)[0]+'_b.fits',clobber=True)
	

if __name__=="__main__":
	jpgToFITS(r'E:\work\workSlumping\dusttest\IMG_0347.jpg')


import numpy as np
from matplotlib import pyplot as plt
import math
from scipy import interpolate as ip
from numpy import ma
import os
import pdb
from copy import deepcopy
from pyGeneralRoutines.span import span
from matplotlib import pyplot as plt
import numpy as np
import os
from pySurf import points as P

method='linear' #default method for interpolation. It can be redefined with Points.method.
#if set differently in function calls, it is overridden.
"""
Finally a point class, inherited from numpy array. 
Points are in format (Npoints,Ndim), where the last column is intended to be the data
    and the first N-1 are the points coordinate. This in the most general case, but routines
    are written and tested on the 3d case. The final intent is make it valid for any ndim
    for an array of 2d shape. Routines made for ndimensional case are marked with ND, routines 
    that need adjustment are marked as 3D.
"""

"""angpos copiata qui, che non so come si fa ad importare in questo casino di directories. 
CHANGES: angpos e' stata corretta funzionava con dati trasposti"""
"""2014/03/17 moved here all routines from EA2019_align and from OP1S07_align."""

"""2014/03/06 from v1. Start from nanovea txt files use pure python to avoid coordinates mess. Assume data are saved with y fast axis and +,+ scanning
directions. It is not guaranteed to work with different scanning directions."""

def _angpos2(xy):
    '''given an image and some notable points, return the angular positions of points with respect to barycenter.
    The angle returned is in the range [-pi:pi]'''

    #N xy points are expressed as array N x 2
    b=(np.sum(xy,0))/(xy.shape[0])    #calculate barycenter
    deltaxy=xy-(b.repeat(xy.size/2).reshape(2,-1)).T
    r=np.sqrt(np.sum(deltaxy**2,1))
    theta=np.arctan2(deltaxy[:,1],deltaxy[:,0])
    return theta,r,b
    
class Points(object):
    """Represent a set of points in a N-dimensional point cloud, defined as array in format (Npoints,Ndim)."""
	#see also version in notebook where is subclassed from array.
    
    def load(self,filename,delimiter='',*args,**kwargs):
        pts=P.get_points(filename,delimiter=delimiter,*args,**kwargs)
        self.points=pts
        return pts

    def __init__(self,filename=None,*args,**kwargs):
        if filename is None:
            return self
        else:
            self.load(filename,*args,**kwargs)
            
    def translate(self,offset=None):
        """returns translated coordinates of 2D point(s) x ([Npoints x 2]) by an offset.
        It works also on 3-D points, in that case the z is returned unchanged."""
        
        #2014/11/26 in class. Replaced old translate with a simple sum of arrays.
        #   The sum is broadcasted whenever it makes sense.
        #It is interesting to notice that calculating the new vector and assigning it to self
        #    with self[:]=translated and returning self
        #    doesn't work, in the sense that the original array is modified and all returned
        #    values point to the same array.
        #2014/08/11 
        #- return a 3d array if a 3d array was passed as argument.
        #- convert the result to array before return.
        #   The result was initially of type np.matrix to simplify operations on extracted columns,
        #   but it didn't really work. The problem is that the result is a matrix and if the
        #   user is unaware of that, all following slicing (column extractions) will be messed up
        #   (vectors will be column vectors as opposite to row vectors returned by np arrays.
        
        ndim=self.points.shape[-1]    
        if offset is None:
            offset=np.zeros(ndim)
        else:
            offset=np.array(offset)
            if np.size(offset)==1:
                offset=np.repeat(offset,ndim)
            elif np.size(offset)!=ndim:
                raise ValueError
                
        self.points=self.points+offset
        return self.points
        
        
    def rotate(self,theta,center=(0,0)):
        """returns rotated coordinates of 2D point(s) x ([Npoints x 2]) about a center with anticlockwise angle theta in rad. 
        If 3D points are passed, z coordinate is maintained."""
        pts=P.rotate_points(self.points,theta, center)
        self.points=pts
        return pts
    
    def resample(self,positions):
        """resample points [Npoints x 3] on the points defined in positions [Mpoints x 2], or [Mpoints x 3]
        (in this case 3rd column is ignored).
        Return a [Nx x Ny , 3] vector of points. To get a (plottable) matrix of data use:
        plt.imshow(rpoints[:,2].reshape(xgrid.size,ygrid.size))."""
        assert self.shape[1]==3
        z=ip.griddata(self[:,0:2],self[:,-1],positions[:,0:2],method=method)
        rpoints=np.hstack([positions[:,0:2],z[:,np.newaxis]])
        return rpoints

    def resample_grid(self,xgrid,ygrid):
        from matplotlib.mlab import griddata
        """resample points [Npoints x 3] on the grid defined by two vectors xgrid [Nx] and ygrid [Ny].
        Return a [Nx * Ny , 3] vector of points, sorted in standard python order
        (x changes faster). To get a (plottable) matrix of data use:
        plt.imshow(rpoints[:,2].reshape(ygrid.size,xgrid.size))."""
        assert self.shape[1]==3
        x,y=np.meshgrid(xgrid,ygrid) #this always return
        z=ip.griddata(self[:,0:2],self[:,2],(x,y),method=method) 
        rpoints=np.vstack([x.flatten(),y.flatten(),z.flatten()]).T   
        #2014/11/24 rpoints=np.vstack([x.T.flatten(),y.T.flatten(),z.T.flatten()]).T
        return rpoints
        
    def save(self,filename,xgrid=None,ygrid=None,shape=None,matrix=False,fill_value=np.nan,**kwargs):
        """save points on a file. If matrix is true write in matrix form (in this case you have to 
        provide the values for axis). Otherwise write as points in columns."""
        #2014/08/08 default fill_value modified to nan.
        #20140420 moved filename to second argument for consistency.
        if matrix:
            if shape==None:
                assert len(xgrid.shape)==len(ygrid.shape)==1
            else:
                assert xgrid==ygrid==None
                x,y,z=np.hsplit(self,3)
                xgrid=np.linspace(x.min(),x.max(),shape[0])
                ygrid=np.linspace(y.min(),y.max(),shape[1])
            assert xgrid!=None
            assert ygrid!=None
            grid=np.vstack([g.flatten() for g in  np.meshgrid(xgrid,ygrid)]).T
            points=ip.griddata(self[:,0:2],self[:,2],grid,fill_value=fill_value,method=method)
            points=self.reshape(ygrid.size,xgrid.size)
            #points=np.hstack([grid,self[:,np.newaxis]])
        #if not, they are already in the correct format
        np.savetxt(self,filename,**kwargs)

    def plot(self,xgrid=None,ygrid=None,shape=None,resample=True,scatter=False,**kwargs):
        """resample xyz points [Nx3] to a grid whose axis xgrid and ygrid are given
        and plot it. If resample is set to False x and y positions are considered only for range, 
        but they are not used to position the z values (it works if x and y are on an exact unrotated grid,
        resampling is slower, but exact)."""
        
        #plot
        #plt.clf()
        x,y,z=np.hsplit(self,3)
        cmap=kwargs.pop('cmap','jet')
        aspect=kwargs.pop('aspect','equal')
        #pdb.set_trace()
        if scatter: #do scatterplot
            beamsize=20 #I may use this to represent beamsize (round symbol) or lateral resolution (square)
            plt.gca().set_aspect(aspect)
            plt.scatter(x, y, c=z, s=beamsize, cmap=cmap, edgecolors='None', **kwargs)
        else:
            #if not scatter plot, grid must be provided as shape or as xgrid and ygrid axis.
            if shape==None:
                if not len(xgrid.shape)==len(ygrid.shape)==1:
                    return z    #skip plot
            else:
                assert xgrid==ygrid==None
                xgrid=np.linspace(x.min(),x.max(),shape[0])
                ygrid=np.linspace(y.min(),y.max(),shape[1])
            if resample:
                print("resampling...")
                z=resample_grid(self,xgrid,ygrid)[:,2]
            nx,ny=[xx.size for xx in (xgrid,ygrid)]
            xr,yr=[span(xx) for xx in (xgrid,ygrid)]
            xxg,sx=np.linspace(xr[0],xr[1],nx,retstep=1)
            yyg,sy=np.linspace(yr[0],yr[1],ny,retstep=1)
            #ranges for plot (intervals centered on xxg, yyg)
            xr=xr+np.array((-sx/2,sx/2))
            yr=yr+np.array((-sy/2,sy/2))
            
            z=z.reshape(ny,nx)
            plt.imshow(z,extent=[xr[0],xr[1],yr[0],yr[1]],interpolation='none',aspect=aspect,
                origin='lower')
        plt.xlabel('X (mm)')
        plt.ylabel('Y (mm)')
        plt.colorbar()
        plt.show()
        return z

    def rms(self,xcrop=None,ycrop=None,mask=None,clip=None):
        """return the rms of the selected points."""
        
        """Note that if instead of deepcopy is used:
        pp=self[:]
        the clipping operation on p2 affects the values of points.
        """
        pp=deepcopy(self)
        z=pp[:,2]
        zm=np.sqrt(np.nanmean(z**2))
        if clip != None:
            if clip[0]!=None:
                z[np.where(z<=clip[0])]=zm
            if clip[1]!=None:
                z[np.where(z>=clip[1])]=zm
        return np.sqrt(np.nanmean(z**2))
        

    def crop(self, xrange=None,yrange=None):
        """crop a xyz points [Nx3], keeping only points inside xrange and yrange defined as (min,max)."""
        if xrange != None: 
            points=points[np.where(xrange[0]<points[:,0]),:][0]
            points=points[np.where(points[:,0]<xrange[1]),:][0]
        if yrange != None: 
            points=points[np.where(yrange[0]<points[:,1]),:][0]
            points=points[np.where(points[:,1]<yrange[1]),:][0]
        return points
        
    def subtract_points(p1,p2,xySecond=False):
        """Subtract second set of points after interpolation on first set coordinates.
        If xySecond is set to True data are interpolate on xy of p2 and then subtracted."""
        if not(xySecond):
            p2=resample_points(p2,p1)
        else:
            p1=resample_points(p1,p2)
        p1[:,2]=p1[:,2]-p2[:,2]
        return p1
        
        
    def smooth_points(self,xywidth):
        """"""
        def gauss_kern(size, sizey=None):
            """ Returns a normalized 2D gauss kernel array for convolutions """
            size = int(size)
            if not sizey:
                sizey = size
            else:
                sizey = int(sizey)
            x, y = mgrid[-size:size+1, -sizey:sizey+1]
            g = exp(-(x**2/float(size)+y**2/float(sizey)))
            return g / g.sum()

        def blur_image(im, n, ny=None) :
            """ blurs the image by convolving with a gaussian kernel of typical
                size n. The optional keyword argument ny allows for a different
                size in the y direction.
            """
            g = gauss_kern(n, sizey=ny)
            improc = signal.convolve(im,g, mode='valid')
            return(improc)
        
        
        x,y,z=np.hsplit(self,3)
        if shape==None:
            assert len(xgrid.shape)==len(ygrid.shape)==1
        else:
            assert xgrid==ygrid==None
            xgrid=np.linspace(x.min(),x.max(),shape[0])
            ygrid=np.linspace(y.min(),y.max(),shape[1])
        if resample:
            print("resampling...")
            z=resample_grid(self,xgrid,ygrid)[:,2]
        z=z.reshape(xgrid.size,ygrid.size)
        z=blur_image(z,xwidth,ywidth)
        plt.imshow(z,extent=[ygrid.min(),ygrid.max(),xgrid.max(),xgrid.min()],aspect=aspect,interpolation='none')
        plt.xlabel('X (mm)')
        plt.ylabel('Y (mm)')
        plt.colorbar()
        plt.show()
        return z
    

if __name__ == "__main__":
    import doctest
    doctest.testmod()